/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/


#include <pdp/pdp_iv.h>
#include <pdp/sched_proc.h>
#include <ta/taiv_dialog.h>

#include <ta/enter_iv.h>
#include <InterViews/window.h>
#include <InterViews/layout.h>
#include <IV-look/kit.h>
#include <InterViews/event.h>
#include <ta/leave_iv.h>

//////////////////////////////////
//	taivSpecMember		//
//////////////////////////////////

int taivSpecMember::BidForMember(MemberDef* md, TypeDef* td) {
  if((td->InheritsFrom(TA_BaseSpec) || td->InheritsFrom(TA_BaseSubSpec))
     && !(md->HasOption("NO_INHERIT")))
    return (taivMember::BidForMember(md,td) + 1);
  return 0;
}

// put some check here for if the spec base has no spec parent
//  (((TA_BaseSpec *) dlg->cur_base)->owner->GetOwner(dlg->typ) == NULL))

//   if((dlg != NULL) && (dlg->cur_base != NULL) && (dlg->typ != NULL) &&
//      dlg->typ->InheritsFrom(TA_BaseSpec) &&
//      (((BaseSpec*)dlg->cur_base)->GetOwner(&TA_BaseSpec) == NULL))
//   {

bool taivSpecMember::NoCheckBox(taivDialog* dlg) {
  if((dlg == NULL) || (dlg->cur_base == NULL) || (dlg->typ == NULL))
    return true;

  if(dlg->typ->InheritsFrom(TA_BaseSpec_MGroup))
    return false;		// always use a check box for these..

  if(dlg->typ->InheritsFrom(TA_BaseSpec)) {
    BaseSpec* bs = (BaseSpec*)dlg->cur_base;
    if(bs->FindParent() != NULL)
      return false;		// owner has a parent spec, needs a box
    return true;		// no owner, no box..
  }
  else if(dlg->typ->InheritsFrom(TA_BaseSubSpec)) {
    BaseSubSpec* bs = (BaseSubSpec*)dlg->cur_base;
    if(bs->FindParent() != NULL)
      return false;		// owner has a parent spec, needs a box
    return true;		// no owner, no box..
  }
  else {
    TAPtr tap = (TAPtr)dlg->cur_base;
    BaseSpec* bs = (BaseSpec*)tap->GetOwner(&TA_BaseSpec); // find an owner..
    if(bs == NULL)
      return true;		// no owner, no box..
    if(bs->FindParent() != NULL)
      return false;		// owner is owned by a spec, needs a box..
  }
  return true;			// default is to not have box...
}


taivData* taivSpecMember::GetDataRep(ivInputHandler* ih, taivDialog* dlg) {
  taivData* rdat;
  if(sub_types != NULL)
    rdat = sub_types->GetDataRep(ih, dlg);
  else
    rdat = taivMember::GetDataRep(ih,dlg);
  if(NoCheckBox(dlg)) {
    return rdat;
  }
  else {
    taivPlusToggle* pt = new taivPlusToggle(rdat, dlg);
    return pt;
  }
}

void taivSpecMember::GetImage(taivData* dat, void* base, ivWindow* win) {
  taivDialog* dlg = dat->dialog;
  if(NoCheckBox(dlg)) {
    if(sub_types != NULL)
      sub_types->GetImage(dat,base,win);
    else
      taivMember::GetImage(dat, base, win);
  }
  else {
    taivPlusToggle* rval = (taivPlusToggle*)dat;
    if(sub_types != NULL)
      sub_types->GetImage(rval->data,base,win);
    else
      taivMember::GetImage(rval->data, base, win);
    bool uniq = false;
    if(typ->InheritsFrom(TA_BaseSpec)) {
      BaseSpec* es = (BaseSpec*)base;
      uniq = es->GetUnique(mbr->idx);
    }
    else {
      BaseSubSpec* es = (BaseSubSpec*)base;
      uniq = es->GetUnique(mbr->idx);
    }
    rval->GetImage(uniq);
    if(uniq)
      rval->orig_val = "true";
    else
      rval->orig_val = "false";
  }
}

void taivSpecMember::GetMbrValue(taivData* dat, void* base, bool& first_diff) {
  taivDialog* dlg = dat->dialog;
  if(NoCheckBox(dlg)) {
    if(sub_types != NULL)
      ((taivMember*)sub_types)->GetMbrValue(dat, base, first_diff);
    else
      taivMember::GetMbrValue(dat, base, first_diff);
    return;
  }

  taivPlusToggle* rval = (taivPlusToggle*)dat;
  if(typ->InheritsFrom(TA_BaseSpec)) {
    BaseSpec* es = (BaseSpec*)base;
    es->SetUnique(mbr->idx,rval->GetValue());
//    es->UpdateMbr(mbr->idx);
  }
  else {
    BaseSubSpec* es = (BaseSubSpec*)base;
    es->SetUnique(mbr->idx,rval->GetValue());
//    es->UpdateMbr(mbr->idx);
  }

  CmpOrigVal(dat, base, first_diff);
  if(sub_types != NULL)
    ((taivMember*)sub_types)->GetMbrValue(rval->data, base, first_diff);
  else
    taivMember::GetMbrValue(rval->data, base, first_diff);
}

void taivSpecMember::CmpOrigVal(taivData* dat, void* base, bool& first_diff) {
  if((taivMisc::record_script == NULL) || !typ->InheritsFrom(TA_taBase))
    return;
  String new_val;
  bool uniq = false;
  if(typ->InheritsFrom(TA_BaseSpec)) {
    BaseSpec* es = (BaseSpec*)base;
    uniq = es->GetUnique(mbr->idx);
  }
  else {
    BaseSubSpec* es = (BaseSubSpec*)base;
    uniq = es->GetUnique(mbr->idx);
  }
  if(uniq)
    new_val = "true";
  else
    new_val = "false";
  if(dat->orig_val == new_val)
    return;
  if(first_diff)
    StartScript(base);
  first_diff = false;

  *taivMisc::record_script << "  ths->SetUnique(\"" << mbr->name << "\", "
			    << new_val << ");\n";
}


//////////////////////////////////
//     ProcessDialog		//
//////////////////////////////////

void ProcessDialog::Ok() {
  Process* prc = (Process*)cur_base;
  if(prc->ctrl_panel.ctrl_panel == this) {
    prc->ctrl_panel.ctrl_panel = NULL;
    prc->ctrl_panel.active = false;
  }
  if(prc->running)
    prc->Stop();
  taivEditDialog::Ok();
}

void ProcessDialog::Cancel() {
  Process* prc = (Process*)cur_base;
  if(prc->ctrl_panel.ctrl_panel == this) {
    prc->ctrl_panel.ctrl_panel = NULL;
    prc->ctrl_panel.active = false;
  }
  if(prc->running)
    prc->Stop();
  taivEditDialog::Cancel();
}

int taivProcess::BidForEdit(TypeDef* td){
  if(td->InheritsFrom(&TA_Process))
    return taivEdit::BidForEdit(td) +1;
  return 0;
}

int taivProcess::Edit(void* base, ivWindow* win, bool wait,bool readonly) {
  ProcessDialog* dlg = (ProcessDialog*)taivMisc::FindEdit(base, typ);
  if(wait || (dlg == NULL) || dlg->CtrlPanel()) {
    dlg = new ProcessDialog(typ, base);
    dlg->cancel_only = readonly;
    dlg->Constr(win, wait);
    return dlg->Edit();
  }
  if(!dlg->waiting) {
    ((NoBlockDialog*)dlg->dialog)->win->deiconify();
    ((NoBlockDialog*)dlg->dialog)->win->raise();
  }
  return 2;
}

void taivProcess::Initialize()	{
  run_iv = NULL;
}

void taivProcess::Destroy() {
  if(run_iv) delete run_iv;
  run_iv = NULL;
}

bool Process_RunDlg::ShowMember(MemberDef* md) {
  if((ProcessDialog::ShowMember(md)) &&
     (md->type->InheritsFrom(&TA_Counter) || md->HasOption("CONTROL_PANEL")))
    return true;
  return false;
}

int taivProcessRunBox::Edit(void* base, ivWindow* win, bool wait,bool readonly) {
  ProcessDialog* dlg = (ProcessDialog*)taivMisc::FindEdit(base, typ);
  if(wait || (dlg == NULL) || !dlg->CtrlPanel()) {
    dlg = new Process_RunDlg(typ, base);
    dlg->cancel_only = readonly;
    dlg->Constr(win, wait);
    return dlg->Edit();
  }
  if(!dlg->waiting) {
    ((NoBlockDialog*)dlg->dialog)->win->deiconify();
    ((NoBlockDialog*)dlg->dialog)->win->raise();
  }
  return 2;
}


//////////////////////////////////
//     StatNewDialog		//
//////////////////////////////////

String_Array StatNewDialog::def_vals;

StatNewDialog::StatNewDialog(Process_MGroup* proc_gp, TypeDef* td, int n_els) {
  procs = proc_gp;
  num = n_els;
  if(td == NULL)
    typ = &TA_Stat;
  else
    typ = td;
  in_proc = (SchedProcess*)proc_gp->DefaultEl(); // this will be set to auto default
  loc = 0;

  if(def_vals.size == 1) {
    loc = (int)def_vals[1];
  }

  num_rep = NULL;
  typ_rep = NULL;
  in_rep = NULL;
  loc_rep = NULL;
  fun_list  = NULL;
}
StatNewDialog::~StatNewDialog() {
  delete num_rep;
  delete typ_rep;
  delete in_rep;
  delete loc_rep;
  delete fun_list;
}
void StatNewDialog::Constr_Box() {
  num_rep = new taivIncrField(String(num), this,true);
  typ_rep = new gpivElTypes(taivMenu::menubar, taivMenu::small,
			       &TA_Stat, typ, this);
  typ_rep->GetMenu();
  typ_rep->GetImage(typ);

  in_rep = new gpivGroupEls(taivMenu::menubar, taivMenu::small, procs,
			  this, false,true,true);
  in_rep->GetMenu();
  GetDefaultProc();
  in_rep->GetImage(procs, (TAPtr)in_proc);

  loc_rep = new taivMenu(taivMenu::menubar,taivMenu::radio_update,
			 taivMenu::small);
  loc_rep->AddItem("DEFAULT");
  loc_rep->AddItem("LOOP");
  loc_rep->AddItem("FINAL");
  loc_rep->GetImage(loc);

  fun_list = new gpivNewFuns(&TA_Stat, this);

  sub_box = layout->hbox
    (taivM->top_spc(layout->vbox(wkit->label("Number:"),
				  taivM->vspc, num_rep->GetLook())),
     taivM->hspc,
     taivM->top_spc(layout->vbox(wkit->label("Of Type:"),
				  taivM->vspc, typ_rep->GetLook())),
     taivM->top_spc(layout->vbox(wkit->label("In Process:"),
				  taivM->vspc, in_rep->GetLook())),
     taivM->top_spc(layout->vbox(wkit->label("Loop/Final:"),
				  taivM->vspc, loc_rep->GetLook())));

  box = layout->vbox(sub_box, taivM->vspc, fun_list->GetRep());
}

void StatNewDialog::GetDefaultProc() {
  in_proc = SchedProcess::Default_StatProc(typ, procs);
}

void StatNewDialog::Constr_Strings(const char*, const char* win_title) {
  taivDialog::Constr_Strings("New Stat: Create new statistic object(s)", win_title);
}

void StatNewDialog::Constr_Handlers() {
  num_rep->SetHandler(dialog);
}

void StatNewDialog::GetValue() {
  num = atoi(num_rep->GetValue());
  typ = typ_rep->GetValue();
  in_proc = (SchedProcess*)in_rep->GetValue();
  loc = loc_rep->GetValue()->itm_no;

  def_vals.EnforceSize(1);
  def_vals[0] = String(loc);
}

Stat* StatNewDialog::New(Process_MGroup* proc_gp, TypeDef* td, ivWindow* win,
		       int n_els)
{
  StatNewDialog* create = new StatNewDialog(proc_gp, td, n_els);
  create->Constr(win, true);
  bool ok_can = create->Edit();
  if(!ok_can || (create->num <= 0) || (create->in_proc == NULL)) {
    delete create;
    return NULL;
  }

  Stat_Group* stat_gp;
  if(create->loc == 0)
    stat_gp = SchedProcess::Default_StatGroup(create->typ, create->in_proc);
  else if(create->loc == 1)
    stat_gp = &(create->in_proc->loop_stats);
  else
    stat_gp = &(create->in_proc->final_stats);

  Stat* rval = (Stat*) stat_gp->New(create->num, create->typ);
  if(taivMisc::record_script != NULL) {
    *taivMisc::record_script << stat_gp->GetPath() << "->New(" << create->num
			      << ", " << create->typ->name << ");\n";
  }
  int i;
  for(i = stat_gp->size -1;i >= (int) (stat_gp->size - create->num); i--){
    create->fun_list->CallFuns(stat_gp->FastEl_(i));
  }
  create->in_proc->UpdateAfterEdit();	// update the process..
  bool auto_edit = taMisc::auto_edit;
  if(create->mouse_button == ivEvent::middle) // reverse of default
    auto_edit = !auto_edit;
  if(create->mouse_button==ivEvent::right) // always edit on rt
    auto_edit = true;
  if(auto_edit) {
    if(create->num == 1) {	// edit-after-create
      rval->Edit(false);
    }
    else {			// edit the group
      stat_gp->Edit(false);
    }
  }
  delete create;
  return rval;
}

//////////////////////////////////
//     ProcNewDialog		//
//////////////////////////////////

String_Array ProcNewDialog::def_vals;

ProcNewDialog::ProcNewDialog(Process_MGroup* proc_gp, TypeDef* td, int n_els) {
  procs = proc_gp;
  num = n_els;
  if(td == NULL)
    typ = &TA_Stat;
  else
    typ = td;
  in_proc = (SchedProcess*)proc_gp->DefaultEl();
  loc = 1;

  if(def_vals.size == 2) {
    loc = (int)def_vals[0];
    MemberDef* md;
    TAPtr bs = tabMisc::root->FindFromPath(def_vals[1], md);
    if(bs != NULL)
      in_proc = (SchedProcess*)bs;
  }

  num_rep = NULL;
  typ_rep = NULL;
  in_rep = NULL;
  loc_rep = NULL;
  fun_list  = NULL;
}
ProcNewDialog::~ProcNewDialog() {
  delete num_rep;
  delete typ_rep;
  delete in_rep;
  delete loc_rep;
  delete fun_list;
}
void ProcNewDialog::Constr_Box() {
  num_rep = new taivIncrField(String(num), this,true);
  typ_rep = new gpivElTypes(taivMenu::menubar, taivMenu::small,
			       &TA_Stat, typ, this);
  typ_rep->GetMenu();
  typ_rep->GetImage(typ);

  in_rep = new gpivGroupEls(taivMenu::menubar, taivMenu::small, procs,
			  this, false,true,true);
  in_rep->GetMenu();
  in_proc = (SchedProcess*)procs->Leaf(0);
  in_rep->GetImage(procs, (TAPtr)in_proc);

  loc_rep = new taivMenu(taivMenu::menubar,taivMenu::radio_update,
			 taivMenu::small);
  loc_rep->AddItem("INIT");
  loc_rep->AddItem("LOOP");
  loc_rep->AddItem("FINAL");
  loc_rep->GetImage(loc);

  fun_list = new gpivNewFuns(&TA_Process, this);

  sub_box = layout->hbox
    (taivM->top_spc(layout->vbox(wkit->label("Number:"),
				  taivM->vspc, num_rep->GetLook())),
     taivM->hspc,
     taivM->top_spc(layout->vbox(wkit->label("Of Type:"),
				  taivM->vspc, typ_rep->GetLook())),
     taivM->top_spc(layout->vbox(wkit->label("In Process:"),
				  taivM->vspc, in_rep->GetLook())),
     taivM->top_spc(layout->vbox(wkit->label("Init/Loop/Final:"),
				  taivM->vspc, loc_rep->GetLook())));

  box = layout->vbox(sub_box, taivM->vspc, fun_list->GetRep());
}

void ProcNewDialog::Constr_Strings(const char*, const char* win_title) {
  taivDialog::Constr_Strings("New Process: Create new process object(s)", win_title);
}

void ProcNewDialog::Constr_Handlers() {
  num_rep->SetHandler(dialog);
}

void ProcNewDialog::GetValue() {
  num = atoi(num_rep->GetValue());
  typ = typ_rep->GetValue();
  in_proc = (SchedProcess*)in_rep->GetValue();
  loc = loc_rep->GetValue()->itm_no;

  def_vals.EnforceSize(2);
  def_vals[0] = String(loc);
  def_vals[1] = TA_SchedProcess_ptr.GetValStr(&in_proc);
}

Process* ProcNewDialog::New(Process_MGroup* proc_gp, TypeDef* td, ivWindow* win,
		       int n_els)
{
  ProcNewDialog* create = new ProcNewDialog(proc_gp, td, n_els);
  create->Constr(win, true);
  bool ok_can = create->Edit();
  if(!ok_can || (create->num <= 0) || (create->in_proc == NULL)) {
    delete create;
    return NULL;
  }

  Process_Group* prgp;
  if(create->loc == 0)
    prgp = &(create->in_proc->init_procs);
  else if(create->loc == 1)
    prgp = &(create->in_proc->loop_procs);
  else
    prgp = &(create->in_proc->final_procs);

  Process* rval = (Process*) prgp->New(create->num, create->typ);
  if(taivMisc::record_script != NULL) {
    *taivMisc::record_script << prgp->GetPath() << "->New(" << create->num
			      << ", " << create->typ->name << ");\n";
  }
  int i;
  for(i = prgp->size -1;i >= (int) (prgp->size - create->num); i--){
    create->fun_list->CallFuns(prgp->FastEl_(i));
  }
  create->in_proc->UpdateAfterEdit();	// update the process..
  bool auto_edit = taMisc::auto_edit;
  if(create->mouse_button == ivEvent::middle) // reverse of default
    auto_edit = !auto_edit;
  if(create->mouse_button==ivEvent::right) // always edit on rt
    auto_edit = true;
  if(auto_edit) {
    if(create->num == 1) {	// edit-after-create
      rval->Edit(false);
    }
    else {			// edit the group
      prgp->Edit(false);
    }
  }
  delete create;
  return rval;
}

