/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// pdp_iv.h : iv stuff for pdp
  
#ifndef pdp_iv_h
#define pdp_iv_h 1

#include <pdp/base.h>
#include <pdp/spec.h>
#include <ta/taiv_type.h>
#include <ta/taiv_dialog.h>
#include <ta_misc/array_iv.h>

class ProcessDialog : public taivEditDialog {
  // #IGNORE
public:
  ProcessDialog(TypeDef* tp, void* base)	: taivEditDialog(tp,base) { };
  ProcessDialog()				{ };
  ~ProcessDialog()			{ };

  virtual bool	CtrlPanel()	{ return false; }
  void		Ok();		// stop the process when these are hit..
  void		Cancel();
};

class Process_RunDlg : public ProcessDialog {
  // #IGNORE 
public:
  Process_RunDlg(TypeDef* tp, void* base) : ProcessDialog(tp,base) {};
  Process_RunDlg() {};
  ~Process_RunDlg() {};

  bool	CtrlPanel()	{ return true; }

  bool ShowMember(MemberDef* md);
};

class taivProcess : public taivEdit {
  // adds control buttons to the edit window
public:  
  taivProcess*	run_iv;

  int		BidForEdit(TypeDef* td);
  int 		Edit(void* base, ivWindow* win, bool wait=false,bool readonly=false);

  void Initialize();
  void Destroy();
  TAIV_EDIT_INSTANCE(taivProcess, taivEdit);
};


class taivProcessRunBox : public taivProcess {
  // just has the control buttons
public:  
  int		BidForEdit(TypeDef*) 	{ return 0; }
  int 		Edit(void* base, ivWindow* win, bool wait=false, bool readonly=false);
  TAIV_EDIT_INSTANCE(taivProcessRunBox, taivProcess);
};


class taivSpecMember : public taivMember {
  // special for the spec type member (adds the unique box)
public:
  int		BidForMember(MemberDef* md, TypeDef* td);
  taivData*	GetDataRep(ivInputHandler* ih, taivDialog* dlg=NULL);
  void		GetImage(taivData* dat, void* base, ivWindow* win); 
  void		GetMbrValue(taivData* dat, void* base, bool& first_diff); 

  void		CmpOrigVal(taivData* dat, void* base, bool& first_diff);

  virtual bool	NoCheckBox(taivDialog* dlg); // deterimine if check box should be used

  TAIV_MEMBER_INSTANCE(taivSpecMember, taivMember);
};

class 	Process_MGroup;

class StatNewDialog : public taivDialog {
  // #IGNORE dialog for creating new statistics 
public:
  Process_MGroup*	procs;
  SchedProcess*		in_proc;
  int			loc;	// location (0=default, 1=loop, 2=final)
  int			num;
  TypeDef*		typ;
  taivField*		num_rep;
  gpivElTypes* 		typ_rep;
  gpivGroupEls*		in_rep;
  taivMenu*		loc_rep;
  gpivNewFuns*		fun_list;
  ivPolyGlyph*		sub_box;

  static String_Array	def_vals; // default values

  StatNewDialog(Process_MGroup* proc_gp, TypeDef* td, int n_els=1);
  ~StatNewDialog();

  void	Constr_Strings(const char* prompt="", const char* win_title="");
  void	Constr_Box();
  void	Constr_Handlers();
  void	GetValue();

  virtual void	GetDefaultProc(); // get default process for given type of stat

  // this is the function you actually call (it figures out default in_process to use)
  static Stat* 	New(Process_MGroup* proc_gp, TypeDef* td = NULL, ivWindow* win=NULL,
		    int n_els=1);
};

class ProcNewDialog : public taivDialog {
  // #IGNORE dialog for creating new processes
public:
  Process_MGroup*	procs;
  SchedProcess*		in_proc;
  int			loc;	// location (0=init, 1=loop, 2=final)
  int			num;
  TypeDef*		typ;
  taivField*		num_rep;
  gpivElTypes* 		typ_rep;
  gpivGroupEls*		in_rep;
  taivMenu*		loc_rep;
  gpivNewFuns*		fun_list;
  ivPolyGlyph*		sub_box;

  static String_Array	def_vals; // default values

  ProcNewDialog(Process_MGroup* proc_gp, TypeDef* td, int n_els=1);
  ~ProcNewDialog();

  void	Constr_Strings(const char* prompt="", const char* win_title="");
  void	Constr_Box();
  void	Constr_Handlers();
  void	GetValue();

  // this is the function you actually call (it uses previous in_proc, or default of 'Trial')
  static Process*  New(Process_MGroup* proc_gp, TypeDef* td = NULL, ivWindow* win=NULL,
		       int n_els=1);
};

#endif // pdp_iv_h
