/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// stuff to implement pdplog_view graphics

#include <pdp/pdplog_iv.h>
#include <pdp/pdplog.h>
#include <ta_misc/datatable_iv.h>

#include <iv_misc/scrollable.h>
#include <iv_misc/hungrybox.h>
#include <iv_misc/vcrbitmaps.h>
#include <iv_misc/dastepper.h>

#include <ta/enter_iv.h>
#include <OS/math.h>
#include <IV-look/button.h>
#include <IV-look/kit.h>
#include <IV-look/dialogs.h>
#include <InterViews/layout.h>
#include <InterViews/style.h>
#include <InterViews/window.h>
#include <InterViews/patch.h>
#include <InterViews/font.h>
#include <InterViews/bitmap.h>
#include <InterViews/stencil.h>
#include <InterViews/background.h>
#include <InterViews/action.h>
#include <InterViews/session.h>
#include <IV-X11/xwindow.h>		// this is for the window title hack

// for page view
#include <InterViews/page.h>
#include <iv_misc/dragpage.h>
#include <iv_misc/adjustbox.h>
#include <InterViews/debug.h>

#include <ta/leave_iv.h>

//////////////////////////////////
//	LogView IV Stuff	//
//////////////////////////////////

void LogView::SetWinName() {
  if(log_mgr == NULL)    return;
  if(!taMisc::iv_active) return;
  if(window == NULL)    return;
  if(name.contains(GetTypeDef()->name) || name.contains(mgr->name)) {
    int idx = mgr->views.FindLeaf(this);
    name = mgr->name + "_" + String(idx);
  }
  String nw_name = String(ivSession::instance()->classname()) + ": " + mgr->GetPath() +
    "(" + mgr->name + ")" + " Vw: " + name.after("_");
  if(log_mgr->log_file.IsOpen() && !log_mgr->log_file.fname.empty()) {
    nw_name += String(" Fn:") + log_mgr->log_file.fname;
    file_name = log_mgr->log_file.fname;
    mgr->file_name = file_name;
  }
  if(nw_name == win_name) return; // no need to set, same name!
  win_name = nw_name;
  mgr->win_name = nw_name;
  if(window->style() == NULL)
    window->style(new ivStyle(wkit->style()));
  window->style()->attribute("title", win_name);
  window->style()->attribute("name", win_name);
  window->style()->attribute("iconName", name);
  if(window->is_mapped()) {
    ivManagedWindowRep* mwRep = window->rep();
    mwRep->do_set(window, &ivManagedWindowRep::set_name);
    mwRep->do_set(window, &ivManagedWindowRep::set_icon_name);
  }
}


//////////////////////////////////
//	LogView_IV_Impl		//
//////////////////////////////////


LogView_IV_Impl::LogView_IV_Impl(LogView* view){
  log_view = view;
  layout = view->layout;
  wkit = view->wkit;
  ckbox = NULL;
}

LogView_IV_Impl::~LogView_IV_Impl(){}
  

void LogView_IV_Impl::SetToggle(bool onoff){
  if(ckbox) 
    ckbox->state()->set(ivTelltaleState::is_chosen,onoff);
}

void LogView_IV_Impl::ToggleDisplay(){
  if(ckbox != NULL) {
    ckbox->state()->set(ivTelltaleState::is_chosen,log_view->display_toggle);
    if(log_view->window != NULL) log_view->window->repair();
  }
}
  
#include <ta/enter_iv.h>
declareActionCallback(LogView)
implementActionCallback(LogView)
#include <ta/leave_iv.h>

ivGlyph* LogView_IV_Impl::GetLogButtons() {

  ivBitmap* fr_bm = new ivBitmap((void *) fr_bits,fr_width,fr_height);
  ivBitmap* fsr_bm = new ivBitmap((void *) fsr_bits,fsr_width,fsr_height);
  ivBitmap* r_bm = new ivBitmap((void *) r_bits,r_width,r_height);
  ivBitmap* f_bm = new ivBitmap((void *) f_bits,f_width,f_height);
  ivBitmap* fsf_bm = new ivBitmap((void *) fsf_bits,fsf_width,fsf_height);
  ivBitmap* ff_bm = new ivBitmap((void *) ff_bits,ff_width,ff_height);

  ivButton* frev = wkit->push_button(new ivStencil(fr_bm,wkit->foreground()),
				 new ActionCallback(LogView)
				 (log_view,&LogView::View_FR));
  ivTelltaleState* fsrev_state = 
    new ivTelltaleState(ivTelltaleState::is_enabled_visible ||
		      ivTelltaleState::is_choosable);
  ivButton* fsrev = new DAStepper
    (wkit->push_button_look
     (new ivStencil(fsr_bm,wkit->foreground()),fsrev_state),
     wkit->style(),fsrev_state,
     new ActionCallback(LogView)
     (log_view,&LogView::View_FSR));

  ivTelltaleState* rev_state = 
    new ivTelltaleState(ivTelltaleState::is_enabled_visible ||
		      ivTelltaleState::is_choosable);
  ivButton* rev = new DAStepper
    (wkit->push_button_look
     (new ivStencil(r_bm,wkit->foreground()),rev_state),
     wkit->style(),rev_state,
     new ActionCallback(LogView)
     (log_view,&LogView::View_R));

  ivTelltaleState* fwd_state = 
    new ivTelltaleState(ivTelltaleState::is_enabled_visible ||
		      ivTelltaleState::is_choosable);
  ivButton* fwd = new DAStepper
    (wkit->push_button_look
     (new ivStencil(f_bm,wkit->foreground()),fwd_state),
     wkit->style(),fwd_state,
     new ActionCallback(LogView)
     (log_view,&LogView::View_F));

  ivTelltaleState* fsfwd_state = 
    new ivTelltaleState(ivTelltaleState::is_enabled_visible ||
		      ivTelltaleState::is_choosable);
  ivButton* fsfwd = new DAStepper
    (wkit->push_button_look
     (new ivStencil(fsf_bm,wkit->foreground()),fsfwd_state),
     wkit->style(),fsfwd_state,
     new ActionCallback(LogView)
     (log_view,&LogView::View_FSF));

  ivButton* ffwd = wkit->push_button(new ivStencil(ff_bm,wkit->foreground()),
				 new ActionCallback(LogView)
				 (log_view,&LogView::View_FF));

  ckbox = wkit->check_box("Display Toggle",
			  new ActionCallback(LogView)
			  (log_view,&LogView::ToggleDisplay));

  ckbox->state()->set(ivTelltaleState::is_chosen,log_view->display_toggle);

  ivButton* clear_button = 
    wkit->push_button("Clear",
		      new ActionCallback(LogView)
		      (log_view,&LogView::Clear));


  ivButton* init_button = 
    wkit->push_button("Init",
		      new ActionCallback(LogView)
		      (log_view,&LogView::InitDisplay));

  ivButton* update_button = 
    wkit->push_button("Update",
		      new ActionCallback(LogView)
		      (log_view,&LogView::UpdateCallback));


  return layout->hbox(layout->vcenter(ckbox),
		      layout->hglue(),
		      layout->vcenter
		      (layout->hbox
		       (layout->hfixed(frev,VCR_BUTTON_WIDTH),
			layout->hfixed(fsrev,VCR_BUTTON_WIDTH),
			layout->hfixed(rev,VCR_BUTTON_WIDTH),
			layout->hfixed(fwd,VCR_BUTTON_WIDTH),
			layout->hfixed(fsfwd,VCR_BUTTON_WIDTH),
			layout->hfixed(ffwd,VCR_BUTTON_WIDTH))),
		      layout->hglue(),
		      layout->vcenter(taivM->small_button(update_button)),
		      layout->vcenter(taivM->small_button(init_button)),
		      layout->vcenter(taivM->small_button(clear_button)));
}

TextLogView_IV_Impl::TextLogView_IV_Impl(LogView * view)
:LogView_IV_Impl(view)
{
  sb = NULL;
  h_line= NULL;
  patch = NULL;
  scrb = NULL;
  vbox = NULL;
  ovbox = NULL;
  cur_line = NULL;
}

TextLogView_IV_Impl::~TextLogView_IV_Impl(){
  ivResource::unref(sb);	sb = NULL;
  ivResource::unref(h_line);	h_line = NULL;
  ivResource::unref(patch);	patch = NULL;
  ivResource::unref(scrb);	scrb = NULL;
  ivResource::unref(vbox);	vbox = NULL;
  ivResource::unref(ovbox);	ovbox = NULL;
  cur_line = NULL;
}

int TLVMore(int* tlv, int newsize, HungryBox*) {
 return ((TextLogView*) tlv)->ShowMore(newsize);
}


int TLVLess(int* tlv, int newsize, HungryBox*) {
  return ((TextLogView*) tlv)->ShowLess(newsize);
}

ivGlyph* TextLogView_IV_Impl::GetBodyRep() {
  ivFontBoundingBox fbb;
  wkit->font()->font_bbox(fbb);
  font_width = fbb.width();
  float font_height = fbb.font_ascent() + fbb.font_descent();

  sb = new TBHungryBox((int*) log_view,&TLVMore,&TLVLess,(ivGlyphIndex) log_view->view_bufsz);
  ivResource::ref(sb);

  wkit->push_style(taivM->title_style);
  h_line = wkit->label("-----------------------No-Current-Header------------------------");
  ivResource::ref(h_line);
  wkit->pop_style();

  vbox = layout->vbox
    (layout->hflexible
     (layout->hbox(layout->hglue(),h_line,layout->hglue())),
     layout->hflexible(wkit->menu_item_separator_look()),
     layout->vnatural(sb,font_height * 4.0));
  ivResource::ref(vbox);

  scrb = (ivPatch *) new HScrollable
    (layout->vcenter
     (vbox,0),0,0,0,0);
  ivResource::ref(scrb);

  patch = new ivPatch(wkit->inset_frame
		      (layout->hmargin(scrb,5)));
  ivResource::ref(patch);

  ivGlyph* log_buttons = GetLogButtons();

  ovbox = layout->vbox
    (layout->hcenter(layout->hflexible(log_buttons,fil,0),0),
     layout->hcenter(layout->margin(patch,5),0),
     layout->hcenter(wkit->hscroll_bar((Scrollable *) scrb),0));
  ivResource::ref(ovbox);

  return(new ivBackground(ovbox, wkit->background()));
}

ivGlyph* TextLogView_IV_Impl::GetPrintData(){
  if(vbox == NULL) return LogView_IV_Impl::GetPrintData();
  return patch;
}

void TextLogView_IV_Impl::AddToLine(const char* name, ivCoord width){
  if(cur_line == NULL)   cur_line = layout->hbox();
  cur_line->append(layout->hnatural(wkit->label(name), font_width * width));
}

void TextLogView_IV_Impl::AddLine(){
  if(cur_line != NULL)  sb->append(cur_line);
  cur_line = NULL;
  vbox->modified(2);
}

void TextLogView_IV_Impl::RemoveLine(int index){
  if(sb == NULL) return;
  if(index == -1) {
    while(sb->count()>0)  sb->remove(sb->count()-1);
  }
  else if(sb->count() > index) sb->remove(index);
  ivResource::flush();
  vbox->modified(2);
}

void TextLogView_IV_Impl::UpdateDisplay(){
  if(patch == NULL) return;
  patch->reallocate();
  patch->redraw();
}

void TextLogView_IV_Impl::UpdateFromBuffer(){
  if(patch == NULL) return;
  vbox->modified(0);
  ovbox->modified(1);
  log_view->main_box->modified(1);
  scrb->reallocate();
  patch->reallocate();
  scrb->redraw();
  patch->redraw();
}


void TextLogView_IV_Impl::StartHeaderLine(){
  ivResource::unref(h_line);
  h_line = layout->hbox();
  ivResource::ref(h_line);
  wkit->push_style(taivM->title_style);
}  


void TextLogView_IV_Impl::AddToHeaderLine(const char* name, ivCoord width){
  if(h_line == NULL)   StartHeaderLine();
  h_line->append(layout->hnatural(wkit->fancy_label(name),font_width * width));
}

void TextLogView_IV_Impl::EndHeaderLine(){
  wkit->pop_style();
  vbox->replace
    (0,layout->hflexible
     (layout->hbox(h_line,layout->hglue())));
  ivResource::flush();

  // notify the boxes
  vbox->modified(0);
}

// page log view

PageLogView_IV_Impl::PageLogView_IV_Impl(LogView * view)
:LogView_IV_Impl(view)
{
  sb = NULL;
  tmp = NULL;
  h_page= NULL;
  patch = NULL;
  scrb = NULL;
  vbox = NULL;
  ovbox = NULL;
  cur_line = NULL;
}

PageLogView_IV_Impl::~PageLogView_IV_Impl(){
  ivResource::unref(sb);	sb = NULL;
  ivResource::unref(h_page);	h_page = NULL;
  ivResource::unref(patch);	patch = NULL;
  ivResource::unref(scrb);	scrb = NULL;
  ivResource::unref(vbox);	vbox = NULL;
  ivResource::unref(ovbox);	ovbox = NULL;
  cur_line = NULL;
}

int PLVMore(int* PLV, int newsize, HungryBox*) {
 return ((PageLogView*) PLV)->ShowMore(newsize);
}


int PLVLess(int* PLV, int newsize, HungryBox*) {
  return ((PageLogView*) PLV)->ShowLess(newsize);
}

void PLVDropUpdate(int* PLV, int) {
  ((PageLogView*) PLV)->UpdateFromDisplay();
}

ivGlyph* PageLogView_IV_Impl::GetBodyRep() {
  ivFontBoundingBox fbb;
  wkit->font()->font_bbox(fbb);
  font_width = fbb.width();
  float font_height = fbb.font_ascent() + fbb.font_descent();

  // header box
  h_page = new DragPage(wkit,layout,
			new ivBackground(nil, wkit->background()));
  h_page->DropUpdate = PLVDropUpdate;
  h_page->dropupdateobject = (int *) log_view;
  ivGlyph* pagewrap =
    layout->vflexible(layout->vnatural(layout->hflexible(layout->hnatural(h_page,300)),30));
  h_page_wrap = new AdjustWrapper(wkit->inset_frame(pagewrap));
  // default header item
  wkit->push_style(taivM->title_style);
  h_page->append
    (wkit->label("-----------------------No-Current-Header------------------------"));
  wkit->pop_style();
  ivResource::ref(h_page);


  sb = new TBHungryBox((int*) log_view,&PLVMore,&PLVLess,(ivGlyphIndex) log_view->view_bufsz);
  ivResource::ref(sb);

  vbox = layout->vbox(3);
  patch = new ivPatch(NULL);

  AdjustBar* adb = new AdjustBar(vbox,patch,wkit,layout,taivM->style);

  vbox->append(h_page_wrap);
  vbox->append(adb);
  vbox->append(new AdjustWrapper(layout->vnatural(sb,font_height * 4.0)));
//  vbox->append(new AdjustWrapper(sb));
  ivResource::ref(vbox);

  scrb = (ivPatch *) new HScrollable
    (layout->vcenter
     (vbox,0),0,0,0,0);
  ivResource::ref(scrb);

  patch->body(wkit->inset_frame(layout->hmargin(scrb,5)));
  ivResource::ref(patch);

  ivGlyph* log_buttons = GetLogButtons();

  ovbox = layout->vbox
    (layout->hcenter(layout->hflexible(log_buttons,fil,0),0),
     layout->hcenter(layout->margin(patch,5),0),
     layout->hcenter(wkit->hscroll_bar((Scrollable *) scrb),0));

  ivResource::ref(ovbox);

  return(new ivBackground(ovbox, wkit->background()));
}

ivGlyph* PageLogView_IV_Impl::GetPrintData(){
  if(vbox == NULL) return LogView_IV_Impl::GetPrintData();
  return patch;
}

void PageLogView_IV_Impl::AddToLine(const char* name, ivCoord x, ivCoord y){
  if(cur_line == NULL)  
//     cur_line = layout->hbox();
    cur_line = (ivPolyGlyph *) new ivPage
      (wkit->outset_frame(layout->shape_of(h_page_wrap)));
  cur_line->append(layout->hnatural(wkit->label(name), font_width * 8));
  ((ivPage *) cur_line)->move(cur_line->count()-1,x+2,y+2); // plus 2 for header's frame
}

void PageLogView_IV_Impl::AddLine(){
  if(cur_line != NULL)  sb->append(cur_line);
  cur_line = NULL;
//  vbox->modified(2); // this is bad it deletes the allocation table
// since it is called inside an "allocate" (via the hungry hungry hungrybox)
// in an adjustwrappper (vbox), we dont want to wipe the allocations we
// are currently using!
}

void PageLogView_IV_Impl::RemoveLine(int index){
  if(sb == NULL) return;
  if(index == -1) {
   while(sb->count()>0)  {
      sb->remove(sb->count()-1);
    }
  }
  else if(sb->count() > index) sb->remove(index);
  ivResource::flush();
//  vbox->modified(2); // this is bad, see addline above...
}

void PageLogView_IV_Impl::UpdateDisplay(){
  if(patch == NULL) return;
  patch->reallocate();
  patch->redraw();
}

void PageLogView_IV_Impl::UpdateFromBuffer(){
  if(patch == NULL) return;
  vbox->modified(0);
  ovbox->modified(1);
  log_view->main_box->modified(1);
  if(scrb != NULL) scrb->reallocate();
  patch->reallocate();
  if(scrb != NULL) scrb->redraw();
  patch->redraw();
}


void PageLogView_IV_Impl::StartHeaderPage(){
  if(h_page == NULL) h_page = new DragPage(wkit,layout);
  while(h_page->count() > 0) h_page->remove(0);
  wkit->push_style(taivM->title_style);
}  


void PageLogView_IV_Impl::AddToHeaderPage(const char* name, ivCoord x, ivCoord y) {
  if(h_page == NULL)   StartHeaderPage();
  if(name == NULL) // no names 
    h_page->append(wkit->label("-----------------------No-Current-Header------------------------"));
  else {
    h_page->append
      (new DragPageObject
       (h_page,wkit,true,true,
	wkit->outset_frame(layout->hnatural(wkit->fancy_label(name),font_width * 8))));
    h_page->GetPage()->move(h_page->count()-1,x,y);
  }
}


void PageLogView_IV_Impl::EndHeaderPage(){
//  AutoAlign();
  wkit->pop_style();
  ivResource::flush();
  // notify the boxes
  vbox->modified(0);
}

void PageLogView_IV_Impl::GetPosition(int i, ivCoord& x, ivCoord& y){
  ivPage* p = h_page->GetPage();
  p->location(i,x,y);
}

void PageLogView_IV_Impl::AutoLayout(){
  h_page->AutoAlign();
  ((PageLogView *) log_view)->UpdateFromDisplay();
}

void PageLogView_IV_Impl::ToggleEdit(){
  h_page->ToggleEdit();
}


GridLogView_IV_Impl::GridLogView_IV_Impl(LogView* view)
: LogView_IV_Impl(view)
{
  adjuster = NULL;
  data_rep = NULL;
  data_patch = NULL;
}

GridLogView_IV_Impl::~GridLogView_IV_Impl() {
  ivResource::unref(adjuster);
  ivResource::unref(data_rep);
}


// The GLV_Action class is given to the color bar in the
// grid view so that when the user changes the values in
// the colorbar by typing a number or twiddling the steppers,
// the grid view's scale range is set to fixed.

class GLV_Action : public ivAction {
public:
  GridLogView* glv;
  void execute();
  GLV_Action(GridLogView* g) {glv = g;}
  ~GLV_Action() {};
};

void GLV_Action::execute(){
  glv->scale_range.grain = ScaleRange::FIXED;
  glv->scale_range.initial = glv->obj->hcsb->range;
  glv->scale_range.offset  = glv->obj->hcsb->zero;
  // turn off all selected buttons
  ivGlyphIndex c = glv->obj->headerbox->count();
  int i;
  for(i=0;i<c;i++){
    HeaderLabel* hl =  (HeaderLabel*) glv->obj->headerbox->component(i);
    hl->SetState(false);
  }
}

void GridLogView_IV_Impl::CreateAdjuster(){
  adjuster = (ivAction *) new GLV_Action((GridLogView *) log_view);
  ivResource::ref(adjuster);
}

ivGlyph* GridLogView_IV_Impl::GetBodyRep(){
  ((GridLogView *) log_view)->GetDataRep();
  data_patch = new ivPatch(data_rep);
  ivGlyph* log_buttons  = GetLogButtons();
  return( layout->vbox
	  (log_buttons,
	   layout->flexible
	   (layout->natural
	    (wkit->inset_frame(data_patch),120,250),fil,100)));
}

void GridLogView_IV_Impl::ScaleOnAllBlocks(){
  GridLogView* glv = (GridLogView *) log_view;
  int i;
  for(i=0; i<glv->obj->headerbox->count(); i++){
    HeaderLabel* hl = (HeaderLabel *) glv->obj->headerbox->component(i);
    if(hl->index == -1) {// a datatable
      if(hl->dtgvs->display_style >= DA_GridViewSpec::BLOCK) {
	glv->scale_on.Add(hl->dtgvs->name);
	hl->SetState(1,false);
      }
    }
    else {
      DA_GridViewSpec* dagvs = (DA_GridViewSpec *) hl->dtgvs->FastEl(hl->index);
      if(dagvs->display_style >= DA_GridViewSpec::BLOCK) {
	glv->scale_on.Add(dagvs->name);
	hl->SetState(1,false);
      }
    }
  }
}

ivGlyph* GridLogView_IV_Impl::GetPrintData(){
  ivGlyph* print_patch = ((GridLogView *) log_view)->obj->print_patch;
  if(print_patch == NULL) return LogView_IV_Impl::GetPrintData();
  else return print_patch;
}

void GridLogView_IV_Impl::InitDisplay(){
  ((GridLogView *) log_view)->GetDataRep();
  data_patch->body(data_rep);
  data_patch->reallocate();
  data_patch->redraw();
  ivResource::flush();
}


void GridLogView_IV_Impl::UpdateDisplay(){
  ((GridLogView *) log_view)->obj->eventpatch->reallocate();
  ((GridLogView *) log_view)->obj->eventpatch->redraw();
}

//////////////////
// GraphLogView //
//////////////////

GraphLogView_IV_Impl::GraphLogView_IV_Impl(LogView* view)
: LogView_IV_Impl(view) {};



ivGlyph* GraphLogView_IV_Impl::GetBodyRep(){
  GraphLogView* glv = (GraphLogView *) log_view;
  ivGlyph* ed_bod = glv->editor->GetLook();
  ivGlyph* log_buttons = GetLogButtons();
  return (layout->vbox(log_buttons, ed_bod));
}

ivGlyph* GraphLogView_IV_Impl::GetPrintData(){
  GraphLogView* glv = (GraphLogView *) log_view;
  if(glv->editor == NULL) return LogView_IV_Impl::GetPrintData();
  glv->editor->viewer->ClearCachedAllocation(); // recompute size for printing
  return glv->editor->boxpatch;
}

