/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the PDP++ software package.			      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, and modify this software and its documentation    //
// for any purpose other than distribution-for-profit is hereby granted	      //
// without fee, provided that the above copyright notice and this permission  //
// notice appear in all copies of the software and related documentation.     //
//									      //
// Permission to distribute the software or modified or extended versions     //
// thereof on a not-for-profit basis is explicitly granted, under the above   //
// conditions. 	HOWEVER, THE RIGHT TO DISTRIBUTE THE SOFTWARE OR MODIFIED OR  //
// EXTENDED VERSIONS THEREOF FOR PROFIT IS *NOT* GRANTED EXCEPT BY PRIOR      //
// ARRANGEMENT AND WRITTEN CONSENT OF THE COPYRIGHT HOLDERS.                  //
// 									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// pdpshell.h
  
#ifndef pdpshell_h
#define pdpshell_h 1

#include <pdp/netstru.h>
#include <pdp/enviro.h>
#include <pdp/process.h>
#include <ta/tdefaults.h>
#include <ta_misc/colorscale.h>

class TypeDefault_MGroup : public MenuGroup_impl {
public:
  int	Dump_Load_Value(istream& strm, TAPtr par=NULL);
  // reset members before loading..
  
  void	Initialize() 		{ SetBaseType(&TA_TypeDefault); }
  void 	Destroy()		{ };
  TA_BASEFUNS(TypeDefault_MGroup);
};

MenuGroup_of(PDPLog);

class Project : public WinBase {
  // ##EXT_proj ##COMPRESS A Project has everything
public:
  WinGeometry		root_win_pos; // position/size of the root window

  TypeDefault_MGroup    defaults;	
  // #NO_FIND #NO_SAVE data for objects
  BaseSpec_MGroup     	specs;		
  // Specification for objects
  Network_MGroup	networks;	
  // Networks in this project
  Environment_MGroup	environments;	
  // Environments for the networks
  Process_MGroup	processes;	
  // Processes in this project
  PDPLog_MGroup		logs;		
  // Logs for this project
  Script_MGroup		scripts;
  // Scripts for this project

  virtual void	LoadDefaults();
  // load defaults according to root::default_file or precompiled defaults

  void	GetWinPos();

  int	Load(istream& strm, TAPtr par=NULL);
  int	Save(ostream& strm, TAPtr par=NULL, int indent=0);

  void	Initialize();
  void 	Destroy()		{ CutLinks(); }
  void 	InitLinks();
  void	CutLinks();
  void	Copy_(const Project& cp);
  COPY_FUNS(Project, WinBase);
  TA_BASEFUNS(Project);
};

// this assumes that MenuGroup_impl is the base group type for menugroups

class Project_MGroup : public MenuGroup_impl {
public:
  int	Load(istream& strm, TAPtr par=NULL); // call reconnect on nets afterwards
  
  void	Initialize() 		{ SetBaseType(&TA_Project); }
  void 	Destroy()		{ };
  TA_BASEFUNS(Project_MGroup);
};


class PDPRoot : public WinBase {
  // structural root of object heirarchy
public:
  static String		version; 	// #READ_ONLY #SHOW current version number
  String		default_file; 	// default name of defaults file
  Project_MGroup	projects; 	// The projects
  ColorScaleSpec_MGroup colorspecs;	// Color Specs

  bool	ThisMenuFilter(MethodDef* md); // don't include saving and loading..
  void	SetWinName();
  void	GetWindow();		// make an app-window
  void  Settings();		// #MENU #MENU_ON_Object edit global settings;
  void	Info();			// #MENU get information/copyright notice
  void	Quit();			
  // #MENU #CONFIRM #MENU_SEP_BEFORE #NO_REVERT_AFTER quit from software..

  void	SetWinPos(float left=0, float bottom=0, float width=0, float height=0);
  // set current window position (uses project for data..)

  void 	Initialize();
  void	InitLinks();
  void 	Destroy();
  TA_BASEFUNS(PDPRoot);
};


#endif // pdpshell_h
