/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

// datatable.cc

#include <values.h>
#include <ctype.h>
#include <ta_misc/datatable.h>
#include <ta_misc/datagraph.h>
#include <ta_misc/datatable_iv.h>
#include <ta/ta_base_tmplt.h> // for int_Array
#include <ta/enter_iv.h>
#include <InterViews/color.h> // for view specs
#include <ta/leave_iv.h>


//////////////////////////
//	DataArray	//
//////////////////////////

String DataArray_impl::opt_after;

void DataArray_impl::Initialize() {
}

void DataArray_impl::InitLinks() {
  taNBase::InitLinks();
}  

void DataArray_impl::Copy_(const DataArray_impl& cp) {
  disp_opts = cp.disp_opts;
}


//////////////////////////
//	float_RArray	//
//////////////////////////

void float_RArray::InitLinks() {
  float_Array::InitLinks();
  taBase::Own(range, this);
}

void float_RArray::Copy_(const float_RArray& cp) {
  range = cp.range;
}

void float_RArray::Add(const float& it) {
  if(size == 0)
    range.Init(it);		// first one is an 'init'
  else
    UpdateRange(it);
  float_Array::Add(it);
}

void float_RArray::Insert(const float& item, int idx, int n_els) {
  UpdateRange(item);
  float_Array::Insert(item, idx, n_els);
}

void float_RArray::UpdateAllRange() {
  if(size == 0) {
    range.Init(0.0f);
    return;
  }
  range.Init(FastEl(0));
  int i;
  for(i=1; i<size; i++)
    UpdateRange(FastEl(i));
}

float float_RArray::Pop(){
  float result = float_Array::Pop();
  UpdateAllRange();
  return result;
}

bool float_RArray::Remove(uint i, int n){
  bool result = float_Array::Remove(i,n);
  if(result) UpdateAllRange();
  return result;
}

bool float_RArray::Remove(const float& item){
  bool result = float_Array::Remove(item);
//  if(result) UpdateAllRange(); // done in Remove(i,n)
  return result;
}

float float_RArray::MaxVal(int& idx, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  idx = start;
  float rval = FastEl(start);
  int i;
  for(i=start+1;i<end;i++) {
    if(FastEl(i) > rval) {
      idx = i;
      rval = FastEl(i);
    }
  }
  return rval;
}

float float_RArray::AbsMaxVal(int& idx, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  idx = start;
  float rval = fabsf(FastEl(start));
  int i;
  for(i=start+1;i<end;i++) {
    if(fabsf(FastEl(i)) > rval) {
      idx = i;
      rval = fabsf(FastEl(i));
    }
  }
  return rval;
}

float float_RArray::MinVal(int& idx, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  idx = start;
  float rval = FastEl(start);
  int i;
  for(i=start+1;i<end;i++) {
    if(FastEl(i) < rval) {
      idx = i;
      rval = FastEl(i);
    }
  }
  return rval;
}

float float_RArray::Sum(int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++)
    rval += FastEl(i);
  return rval;
}

float float_RArray::Mean(int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end <= start)	return 0.0f;
  int sz = end - start;
  return Sum(start,end) / (float)sz;
}
  
float float_RArray::Var(float mean, bool use_mean, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end <= start)	return 0.0f;
  if(!use_mean)    mean = Mean(start,end);
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++)
    rval += (FastEl(i) - mean) * (FastEl(i) - mean);
  int sz = (end - start)-1;
  if(sz == 0) sz = 1;
  return rval / (float)sz;
}

float float_RArray::StdDev(float mean, bool use_mean, int start, int end) const {
  return sqrtf(Var(mean, use_mean, start, end));
}
  
float float_RArray::SEM(float mean, bool use_mean, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end <= start)	return 0.0f;
  return StdDev(mean, use_mean, start, end) / sqrtf(end - start);
}
  
float float_RArray::Covar(const float_RArray& oth, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  if(end <= start)	return 0.0f;
  float my_mean = Mean(start,end);
  float oth_mean = oth.Mean(start,end);
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++)
    rval += (FastEl(i) - my_mean) * (oth.FastEl(i) - oth_mean);
  int sz = end - start;
  return rval / (float)sz;
}
  
float float_RArray::Correl(const float_RArray& oth, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  if(end <= start)	return 0.0f;
  float my_mean = Mean(start,end);
  float oth_mean = oth.Mean(start,end);
  float my_var = 0.0f;
  float oth_var = 0.0f; 
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++) {
    float my_val = FastEl(i) - my_mean;
    float oth_val = oth.FastEl(i) - oth_mean;
    rval += my_val * oth_val;
    my_var += my_val * my_val;
    oth_var += oth_val * oth_val;
  }
  float var_prod = sqrtf(my_var * oth_var);
  if(var_prod != 0.0f)
    return rval / var_prod;
  else
    return 0.0f;
}

float float_RArray::SSLength(int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++)
    rval += FastEl(i) * FastEl(i);
  return rval;
}

float float_RArray::InnerProd(const float_RArray& oth, bool norm,
			      int start, int end) const
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  if(end == 0)    return 0.0f;
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++)
    rval += FastEl(i) * oth.FastEl(i);
  if(norm) {
    float dist = sqrtf(SSLength(start, end) * oth.SSLength(start, end));
    if(dist != 0.0f)
      rval /= dist;
  }
  return rval;
}

float float_RArray::SumSquaresDist(const float_RArray& oth, bool norm, float tolerance,
				   int start, int end) const
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++) {
    float d = FastEl(i) - oth.FastEl(i);
    if(fabs(d) <= tolerance)  d = 0.0f;
    rval += d * d;
  }
  if(norm) {
    float dist = SSLength(start, end) + oth.SSLength(start, end);
    if(dist != 0.0f)
      rval /= dist;
  }
  return rval;
}

float float_RArray::EuclidDist(const float_RArray& oth, bool norm, float tolerance,
				   int start, int end) const
{
  return sqrtf(SumSquaresDist(oth, norm, tolerance, start, end));
}

float float_RArray::HammingDist(const float_RArray& oth, bool norm, float tolerance,
				   int start, int end) const
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  float rval = 0.0f;
  float alen = 0.0f;
  float blen = 0.0f;
  int i;
  for(i=start;i<end;i++) {
    float d = fabs(FastEl(i) - oth.FastEl(i));
    if(d <= tolerance)  d = 0.0f;
    rval += d;
    if(norm) {
      alen += fabs(FastEl(i));
      blen += fabs(oth.FastEl(i));
    }
  }
  if(norm) {
    float dist = alen + blen;
    if(dist != 0.0f)
      rval /= dist;
  }
  return rval;
}

float float_RArray::CrossEntropy(const float_RArray& oth, int start, int end) const
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  if(end == 0)    return 0.0f;
  float rval = 0.0f;
  int i;
  for(i=start;i<end;i++) {
    float p = FastEl(i);
    float q = oth.FastEl(i);
    q = MAX(q,0.000001f); q = MIN(q,0.999999f);
    if(p >= 1.0f)
      rval += -logf(q);
    else if(p <= 0.0f)
      rval += -logf(1.0f - q);
    else
      rval += p * logf(p/q) + (1.0f - p) * logf((1.0f - p) / (1.0f - q));
  }
  return rval;
}

float float_RArray::Dist(const float_RArray& oth, DistMetric metric, bool norm,
			 float tolerance, int start, int end) const
{
  switch(metric) {
  case SUM_SQUARES:
    return SumSquaresDist(oth, norm, tolerance, start, end);
  case EUCLIDIAN:
    return EuclidDist(oth, norm, tolerance, start, end);
  case HAMMING:
    return HammingDist(oth, norm, tolerance, start, end);
  case COVAR:
    return Covar(oth, start, end);
  case CORREL:
    return Correl(oth, start, end);
  case INNER_PROD:
    return InnerProd(oth, norm, start, end);
  case CROSS_ENTROPY:
    return CrossEntropy(oth, start, end);
  }
  return 0.0f;
}

bool float_RArray::LargerFurther(DistMetric metric) {
  switch(metric) {
  case SUM_SQUARES:
    return true;
  case EUCLIDIAN:
    return true;
  case HAMMING:
    return true;
  case COVAR:
    return false;
  case CORREL:
    return false;
  case INNER_PROD:
    return false;
  case CROSS_ENTROPY:
    return false;
  }
  return 0.0f;
}
  
void float_RArray::Histogram(const float_RArray& oth, float bin_size) {
  Reset();
  float_RArray tmp = oth;	// need to sort it!
  tmp.Sort();
  float min = tmp[0];
  float max = tmp.Peek();
  int src_idx = 0;
  int trg_idx = 0;
  float cur_val;
  for(cur_val = min; cur_val <= max; cur_val += bin_size, trg_idx++) {
    float cur_max = cur_val + bin_size;
    Add(0);
    float& cur_hist = FastEl(trg_idx);
    while((src_idx < tmp.size) && (tmp[src_idx] < cur_max)) {
      cur_hist += 1.0f;
      src_idx++;
    }
  }
}

void float_RArray::AggToArray(const float_RArray& oth, Aggregate& agg,
			      int start, int end)
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  end = MIN(end, oth.size);
  int i;
  for(i=start;i<end;i++)
    agg.ComputeAggNoUpdt(FastEl(i), oth.FastEl(i));
  UpdateAllRange();
}

float float_RArray::AggToVal(Aggregate& agg, int start, int end) const {
  if(end == -1)	end = size;  else end = MIN(size, end);
  float rval = agg.InitAggVal();
  agg.Init();
  int i;
  for(i=start;i<end;i++)
    agg.ComputeAgg(rval, FastEl(i));
  return rval;
}

void float_RArray::CopyVals(const taArray_impl& from, int start, int end, int at) {
  float_Array::CopyVals(from, start, end, at);
  UpdateAllRange();
}
    
float float_RArray::NormLen(float len, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0) 	return 0.0f;
  float scale = (len * len) / SSLength(start,end);
  int i;
  for(i=start;i<end;i++) {
    float mag = (FastEl(i) * FastEl(i)) * scale;
    FastEl(i) = (FastEl(i) >= 0.0f) ? mag : -mag;
  }
  UpdateAllRange();
  return scale;
}

float float_RArray::NormSum(float sum, float min_val, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)	return 0.0f;
  float act_sum = 0.0f;
  int i;
  for(i=start;i<end;i++)
    act_sum += (FastEl(i) - range.min);
  float scale = (sum / act_sum);
  for(i=start;i<end;i++)
    FastEl(i) = ((FastEl(i) - range.min) * scale) + min_val;
  UpdateAllRange();
  return scale;
}

float float_RArray::NormMax(float max, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)	return 0.0f;
  int idx;
  float cur_max = MaxVal(idx, start, end);
  float scale = (max / cur_max);
  int i;
  for(i=start;i<end;i++)
    FastEl(i) *= scale;
  UpdateAllRange();
  return scale;
}

float float_RArray::NormAbsMax(float max, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)	return 0.0f;
  int idx;
  float cur_max = AbsMaxVal(idx, start, end);
  float scale = (max / cur_max);
  int i;
  for(i=start;i<end;i++)
    FastEl(i) *= scale;
  UpdateAllRange();
  return scale;
}

void float_RArray::SimpleMath(const SimpleMathSpec& math_spec, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)  return;
  int i;
  for(i=start;i<end;i++)
    FastEl(i) = math_spec.Evaluate(FastEl(i));
  UpdateAllRange();
}

void float_RArray::SimpleMathArg(const float_RArray& arg_ary, const SimpleMathSpec& math_spec,
				 int start, int end)
{
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)  return;
  SimpleMathSpec myms = math_spec;
  int i;
  for(i=start;i<end;i++) {
    if(i < arg_ary.size)
      myms.arg = arg_ary.FastEl(i);
    FastEl(i) = myms.Evaluate(FastEl(i));
  }
  UpdateAllRange();
}

int float_RArray::Threshold(float thresh, float low, float high, int start, int end) {
  if(end == -1)	end = size;  else end = MIN(size, end);
  if(end == 0)  return 0;
  int rval = 0;
  int i;
  for(i=start;i<end;i++) {
    if(FastEl(i) >= thresh) {
      FastEl(i) = high;
      rval++;
    }
    else
      FastEl(i) = low;
  }
  range.min = low;
  range.max = high;
  return rval;
}

void float_RArray::ShiftLeft(int nshift) {
  float_Array::ShiftLeft(nshift);
  UpdateAllRange();
} 

int float_RArray::Dump_Load_Value(istream& strm, TAPtr par) {
  int rval = float_Array::Dump_Load_Value(strm, par);
  UpdateAllRange();
  return rval;
}

void float_RArray::WritePoint(const TwoDCoord& geom, int x, int y, float color, bool wrap) {
  bool clipped = (TwoDCoord::WrapClipOne(wrap, x, geom.x) ||
		  TwoDCoord::WrapClipOne(wrap, y, geom.y));
  if(!wrap && clipped)
    return;
  El((y * geom.x) + x) = color;
} 

void float_RArray::RenderLine(const TwoDCoord& geom, int xs, int ys, int xe, int ye, 
			      float color, bool wrap)
{
  int xd = xe - xs;
  int yd = ye - ys;
  int x,y;
  if(abs(xd) > abs(yd)) {
    if(yd > 0) yd++;
    else (yd--);
    if(xd >= 0) {
      for(x=xs; x<=xe; x++) {
	y = ys + (yd * (x - xs)) / abs(xd);
	if(yd > 0) y = MIN(ye, y);
	else y = MAX(ye, y);
	WritePoint(geom, x, y, color, wrap);
      }
    }
    else {
      for(x=xs; x>=xe; x--) {
	y = ys + (yd * (xs - x)) / abs(xd);
	if(yd > 0) y = MIN(ye, y);
	else y = MAX(ye, y);
	WritePoint(geom, x, y, color, wrap);
      }
    }
  }
  else {
    if(xd > 0) xd++;
    else (xd--);
    if(yd >= 0) {
      for(y=ys; y<=ye; y++) {
	x = xs + (xd * (y - ys)) / abs(yd);
	if(xd > 0) x = MIN(xe, x);
	else x = MAX(xe, x);
	WritePoint(geom, x, y, color, wrap);
      }
    }
    else {
      for(y=ys; y>=ye; y--) {
	x = xs + (xd * (ys - y)) / abs(yd);
	if(xd > 0) x = MIN(xe, x);
	else x = MAX(xe, x);
	WritePoint(geom, x, y, color, wrap);
      }
    }
  }
}

void float_RArray::WriteXPoints(const TwoDCoord& geom, int x, int y, 
				const float_RArray& color, int wdth, bool wrap)
{
  int del = (wdth - 1) / 2;
  int i;
  for(i=0; i<wdth; i++)
    WritePoint(geom, x, y+i-del, color[i], wrap);
}

void float_RArray::WriteYPoints(const TwoDCoord& geom, int x, int y, 
				const float_RArray& color, int wdth, bool wrap)
{
  int del = (wdth - 1) / 2;
  int i;
  for(i=0; i<wdth; i++)
    WritePoint(geom, x+i-del, y, color[i], wrap);
}

void float_RArray::RenderWideLine(const TwoDCoord& geom, int xs, int ys, int xe, int ye, 
				  const float_RArray& color, int wdth, bool wrap)
{
  int xd = xe - xs;
  int yd = ye - ys;
  int x,y;
  if(abs(xd) > abs(yd)) {
    if(xd >= 0) {
      for(x=xs; x<=xe; x++) {
	y = ys + (yd * (x - xs)) / abs(xd);
	WriteXPoints(geom, x, y, color, wdth, wrap);
      }
    }
    else {
      for(x=xs; x>=xe; x--) {
	y = ys + (yd * (xs - x)) / abs(xd);
	WriteXPoints(geom, x, y, color, wdth, wrap);
      }
    }
  }
  else {
    if(yd >= 0) {
      for(y=ys; y<=ye; y++) {
	x = xs + (xd * (y - ys)) / abs(yd);
	WriteYPoints(geom, x, y, color, wdth, wrap);
      }
    }
    else {
      for(y=ys; y>=ye; y--) {
	x = xs + (xd * (ys - y)) / abs(yd);
	WriteYPoints(geom, x, y, color, wdth, wrap);
      }
    }
  }
}


//////////////////////////
//	DataTable	//
//////////////////////////

void DataTable::Initialize() {
  SetBaseType(&TA_DataArray);	// the impl doesn't inherit properly..
}

void DataTable::Destroy() {
}

void DataTable::InitLinks() {
  taGroup<DataArray_impl>::InitLinks();
}

void DataTable::UpdateAfterEdit() {
  taGroup<DataArray_impl>::UpdateAfterEdit();
}

void DataTable::ResetData() {
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)	ar->NewAR();
    else      			ar->AR()->Reset();
  }
}

void DataTable::RemoveRow(int row_num) {
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)	ar->NewAR();
    else      			ar->AR()->Remove(row_num);
  }
}

void DataTable::AddBlankRow() {
  if(leaves == 0)    return;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)      ar->NewAR();
    else if(ar->InheritsFrom(TA_String_Data))
      ((String_Array*)ar->AR())->Add("");
    else 
      ((float_RArray*)ar->AR())->Add(0);
  }
}

void DataTable::AddRowToArray(float_RArray& tar, int row_num) const {
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    float val = 0;
    if(ar->AR() == NULL)	ar->NewAR();
    else if(ar->InheritsFrom(TA_float_Data))
      val = ((float_RArray*)ar->AR())->El(row_num);
    tar.Add(val);
  }
}

void DataTable::AggRowToArray(float_RArray& tar, int row_num, Aggregate& agg) const {
  if(tar.size < leaves)
    tar.Insert(0, tar.size, leaves - tar.size);
  int cnt = 0;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    float val = 0;
    if(ar->AR() == NULL)	ar->NewAR();
    else if(ar->InheritsFrom(TA_float_Data))
      val = ((float_RArray*)ar->AR())->El(row_num);
    agg.ComputeAggNoUpdt(tar.FastEl(cnt), val);
    cnt++;
  }
}

float DataTable::AggRowToVal(int row_num, Aggregate& agg) const {
  float rval = agg.InitAggVal();
  agg.Init();
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    float val = 0;
    if(ar->AR() == NULL)	ar->NewAR();
    else if(ar->InheritsFrom(TA_float_Data))
      val = ((float_RArray*)ar->AR())->El(row_num);
    agg.ComputeAgg(rval, val);
  }
  return rval;
}

void DataTable::AddArrayToRow(float_RArray& tar) {
  if(tar.size == 0)	return;
  int cnt = 0;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL) ar->NewAR();
    if(ar->InheritsFrom(TA_float_Data))
      ((float_RArray*)ar->AR())->Add(tar.FastEl(cnt));
    cnt++;
    if(cnt >= tar.size)	break;
  }
}

void DataTable::AggArrayToRow(float_RArray& tar, int row_num, Aggregate& agg) {
  if(tar.size == 0)	return;
  int cnt = 0;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)	ar->NewAR();
    if(ar->InheritsFrom(TA_float_Data)) {
      float& val = ((float_RArray*)ar->AR())->El(row_num);
      agg.ComputeAggNoUpdt(val, tar.FastEl(cnt));
    }
    cnt++;
    if(cnt >= tar.size)	break;
  }
}

int DataTable::MaxLength() {
  int max = 0;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)      	ar->NewAR();
    else			max = MAX(max,ar->AR()->size);
  }
  return max;
}    

int DataTable::MinLength() {
  if(size == 0) return 0;
  int min = MAXINT;
  taLeafItr i;
  DataArray_impl* ar;
  FOR_ITR_EL(DataArray_impl, ar, this->, i) {
    if(ar->AR() == NULL)      	ar->NewAR();
    else			min = MIN(min,ar->AR()->size);
  }
  return min;
}

//////////////////////////
// 	ClustNode	//
//////////////////////////

void ClustLink::Initialize() {
  dist = 0.0f;
  node = NULL;
}

void ClustLink::Copy_(const ClustLink& cp) {
  dist = cp.dist;
  taBase::SetPointer((TAPtr*)&node, cp.node);
}

void ClustLink::CutLinks() {
  taBase::DelPointer((TAPtr*)&node);
  taBase::CutLinks();
}

void ClustNode::Initialize() {
  pat = NULL;
  par_dist = 0.0f;
  nn_dist = 0.0f;
  tmp_dist = 0.0f;
  y = 0.0f;
}

void ClustNode::AddChild(ClustNode* nd, float dst) {
  ClustLink* lk = new ClustLink;
  taBase::SetPointer((TAPtr*)&(lk->node), nd);
  lk->dist = dst;
  children.Add(lk);
}

void ClustNode::LinkNN(ClustNode* nd, float dst) {
  ClustLink* lk = new ClustLink;
  taBase::SetPointer((TAPtr*)&(lk->node), nd);
  lk->dist = dst;
  nns.Add(lk);
}

bool ClustNode::RemoveChild(ClustNode* nd) {
  bool rval = false;
  int i;
  for(i=children.size-1;i>=0;i--) {
    if(GetChild(i) == nd) {
      children.Remove(i);
      rval = true;
    }
  }
  return rval;
}

int ClustNode::FindChild(ClustNode* nd) {
  int i;
  for(i=0;i<children.size;i++) {
    if(GetChild(i) == nd)
      return i;
  }
  return -1;
}

void ClustNode::Cluster(float_RArray::DistMetric metric,
			bool norm, float tol)
{
  if(!float_RArray::LargerFurther(metric)) {
    taMisc::Error("Cluster requires distance metric where larger = further apart");
    return;
  }
  do {
    // set nearest neighbor pointers
    NNeighbors(metric, norm, tol);
    // find closest and make a new node
  } while (ClustOnClosest(metric));

  SetParDists(0.0f, metric);
}

void ClustNode::Graph(ostream& strm) {
  SetYs(0.5f);
  nn_dist = .1;
  Graph_impl(strm);
}

void ClustNode::Graph_impl(ostream& strm) {
  if(pat != NULL) {
    strm << par_dist - nn_dist << " " << y << "\n";
    strm << par_dist << " " << y << " \"" << name << "\"\n";
    strm << par_dist - nn_dist << " " << y << "\n";
    strm << par_dist - nn_dist << " " << y << "\n";
  }
  else {
    strm << par_dist - nn_dist << " " << y << "\n";
    strm << par_dist << " " << y << "\n";

    int i;
    for(i=0; i<children.size; i++) {
      ClustLink* nd = (ClustLink*)children[i];
      nd->node->Graph_impl(strm);
      strm << par_dist << " " << y << "\n";
    }
    
    strm << par_dist - nn_dist << " " << y << "\n";
  }
}

void ClustNode::XGraph(const char* fnm, const char* title) {
  fstream fh;
  fh.open(fnm, ios::out);
  fh << "TitleText: " << title << "\n";
  Graph(fh);
  fh.close();
  String cmd = String("xgraph -0 \"\" ") + fnm + "&";
  system(cmd);
}

bool ClustNode::ClustOnClosest(float_RArray::DistMetric metric) {
  if(children.size < 2)
    return false;		// cannot have any more clustering to do!

  float min_d = MAXFLOAT;
  int min_idx=-1;
  int i;
  for(i=0; i<children.size; i++) { // find node with closest neighbors
    ClustNode* nd = GetChild(i);
    if(nd->nn_dist < min_d) {
      min_d = nd->nn_dist;
      min_idx = i;
    }
  }
  if(min_idx < 0) return false;
  // make a new cluster around this node
  ClustNode* nd = GetChild(min_idx);
  ClustNode* new_clust = new ClustNode;
  AddChild(new_clust, min_d);
  // add the min node and its nearest neighbors to the new cluster
  new_clust->AddChild(nd);
  children.Remove(min_idx);
  for(i=0;i<nd->nns.size; i++) {
    ClustLink* nlk = (ClustLink*)nd->nns[i];
    if(nlk->dist != min_d) {
      continue;
    }
    new_clust->AddChild(nlk->node);
    RemoveChild(nlk->node); // and the nns
  }

  // then finally check if any other nns at min_d have other nns not already obtained
  for(i=0;i<nd->nns.size; i++) {
    ClustLink* nlk = (ClustLink*)nd->nns[i];
    if(nlk->dist != min_d)
      continue;
    ClustNode* nn = nlk->node;
    int j;
    for(j=0; j<nn->nns.size; j++) {
      ClustLink* nn_nlk = (ClustLink*)nn->nns[j];
      if(nn_nlk->dist != min_d) {
	continue;
      }
      if(new_clust->FindChild(nn_nlk->node) < 0) { // not in new clust yet
	new_clust->AddChild(nn_nlk->node);
	RemoveChild(nn_nlk->node); // and remove from main list
      }
    }
  }

  if(children.size < 2)
    return false;		// cannot have any more clustering to do!
  return true;
}
  
void ClustNode::NNeighbors(float_RArray::DistMetric metric,
		    bool norm, float tol)
{
  int i;
  for(i=0; i<children.size; i++) {
    ClustNode* nd = GetChild(i);
    nd->nns.Reset();
    nd->nn_dist = MAXFLOAT;
  }
  for(i=0; i<children.size; i++) {
    ClustNode* nd = GetChild(i);
    float min_d = MAXFLOAT;
    int j;
    for(j=i+1; j<children.size; j++) {
      ClustNode* ond = GetChild(j);
      ond->tmp_dist = nd->Dist(*ond, metric, norm, tol);
      min_d = MIN(ond->tmp_dist, min_d);
    }
    for(j=i+1; j<children.size; j++) {
      ClustNode* ond = GetChild(j);
      if(ond->tmp_dist == min_d) {
	nd->LinkNN(ond, min_d);	// link together with distance
	ond->LinkNN(nd, min_d);
      }
    }
  }
  // now make a 2nd pass and get smallest distance for each node and its neighbors
  for(i=0; i<children.size; i++) {
    ClustNode* nd = GetChild(i);
    nd->nn_dist = MAXFLOAT;
    int j;
    for(j=0; j<nd->nns.size; j++) {
      ClustLink* nlk = (ClustLink*)nd->nns[j];
      nd->nn_dist = MIN(nd->nn_dist, nlk->dist);
    }
  }
}

float ClustNode::Dist(const ClustNode& oth, float_RArray::DistMetric metric,
		      bool norm, float tol) const
{
  if(pat != NULL) {
    if(oth.pat != NULL) {
      float dst = pat->Dist(*(oth.pat), metric, norm, tol);
      return dst;
    }
    else {
      float dst = oth.Dist(*this, metric, norm, tol);
      return dst;
    }
  }
  float avg_dist = 0.0f;
  int i;
  for(i=0; i<children.size; i++) {
    ClustNode* nd = GetChild(i);
    if(nd->pat != NULL) {
      if(oth.pat != NULL) {
	float dst = nd->pat->Dist(*(oth.pat), metric, norm, tol);
	avg_dist += dst;
      }
      else {
	float dst = oth.Dist(*nd, metric, norm, tol);
	avg_dist += dst;
      }
    }
    else {
      float dst = nd->Dist(oth, metric, norm, tol);
      avg_dist += dst;
    }
  }
  return avg_dist / (float)children.size;
}

void ClustNode::SetYs(float y_init) {
  static float global_y;
  if(y_init != -1.0f)
    global_y = y_init;
  if(pat == NULL) {
    float y_avg = 0.0f;
    int i;
    for(i=0; i<children.size; i++) {
      ClustNode* nd = GetChild(i);
      nd->SetYs();
      y_avg += nd->y;
    }
    y_avg /= (float)children.size; // average of all y's of kids
    y = y_avg;
  }
  else {
    y = global_y;
    global_y += 1.0f;
  }
}

float ClustNode::SetParDists(float par_d, float_RArray::DistMetric metric) {
  //  bool larger_further = float_RArray::LargerFurther(metric);
  float max_d;
  //  if(larger_further)
    max_d = -MAXFLOAT;
//   else
//     max_d = MAXFLOAT;
  par_dist = par_d + nn_dist;
  int i;
  for(i=0; i<children.size; i++) {
    ClustNode* nd = GetChild(i);
    float dst = nd->SetParDists(par_dist, metric);
    //    if(larger_further)
      max_d = MAX(dst, max_d);
//     else
//       max_d = MIN(dst, max_d);
  }
  return max_d;
}

//////////////////////////
// 	DA View Specs	//
//////////////////////////

void DA_ViewSpec::Initialize(){
  visible = true;
  data_array = NULL;
}

void DA_ViewSpec::CutLinks() {
  taBase::DelPointer((TAPtr*)&data_array);
  taNBase::CutLinks();
}

void DA_ViewSpec::Copy_(const DA_ViewSpec& cp) {
  display_name = cp.display_name;
  visible = cp.visible;
}

void DA_ViewSpec::UpdateAfterEdit() {
  taNBase::UpdateAfterEdit();
  if(data_array == NULL) {
    DT_ViewSpec* dtv = GET_MY_OWNER(DT_ViewSpec);
    if(dtv != NULL) {
      int idx = dtv->Find(this);
      if(idx >= 0) {
	DataTable* dt = dtv->data_table;
	if(dt != NULL) {
	  DataArray_impl* da = dt->El(idx);
	  if(da != NULL)
	    taBase::SetPointer((TAPtr*)&data_array, da);
	}
	else
	  taMisc::Error("*** Null data_table in:", dtv->GetPath());
      }
    }
  }
  if(!(name.empty()) && (display_name.empty()))
    display_name = DA_ViewSpec::CleanName(name);
}

void DA_ViewSpec::UpdateView() {
  DT_ViewSpec* dtv = GET_MY_OWNER(DT_ViewSpec);
  if(dtv != NULL) dtv->UpdateAfterEdit();
}

void DA_ViewSpec::ButtonToggle(){
  visible = !visible;
  UpdateView();
  char vstring = (visible ?  '1' : '0');
  taivMisc::RecordScript(GetPath() + ".visible = " + vstring + ";\n");
  taivMisc::RecordScript(GetPath() + ".UpdateView();\n");
}

int DA_ViewSpec::BuildFromDataArray(DataArray_impl* nda) {
  float first_time = false;
  if(data_array == NULL) first_time = true;
  if(nda != NULL) {
    if(data_array == NULL)
      taBase::SetPointer((TAPtr*)&data_array, nda);
  }
  if(data_array == NULL)
    UpdateAfterEdit();
  if(data_array == NULL)	// still..
    return false;

  // always replace name and throwout old data
  // don't throw out the old data if we are just now creating this
  // view since this results in the user not being able to create
  // an additional view of data in a current view. Since the names
  // will never originally match up, the array's were always being reset.

  if(DA_ViewSpec::CleanName(name) != DA_ViewSpec::CleanName(data_array->name)) {
    if((first_time == false) && (data_array->AR() != NULL))
      data_array->AR()->Reset();
    name = data_array->name;
    display_name = DA_ViewSpec::CleanName(name);
  }
  if(data_array->HasDispOption(" HIDDEN, "))
    visible = false;
  return true;
}


String DA_ViewSpec::CleanName(String& name){
  String result = name;
  // <xxx> is a comment and should be removed
  int pos;
  if(((pos = result.index('<')) != -1) &&
     ((pos = result.index('>',pos+1)) != -1))
    result = result.after(pos);
  // any non-alpha-numeric initial char should be removed
  if((result.length() > 1) && !isalnum(result[0]))
    result = result.after(0);
  return result;
}

//////////////////////////
// 	DT View Specs	//
//////////////////////////

void DT_ViewSpec::Initialize(){
  data_table = NULL;
  visible = true;
}

void DT_ViewSpec::Destroy(){
  CutLinks();
}

void DT_ViewSpec::InitLinks() {
  taGroup<DA_ViewSpec>::InitLinks();
}  

void DT_ViewSpec::CutLinks() {
  taBase::DelPointer((TAPtr*)&data_table);
  taGroup<DA_ViewSpec>::CutLinks();
}

void DT_ViewSpec::Copy_(const DT_ViewSpec& cp) {
  display_name = cp.display_name;
  visible = cp.visible;
}

void DT_ViewSpec::UpdateAfterEdit(){
  taGroup<DA_ViewSpec>::UpdateAfterEdit();
  BuildFromDataTable(data_table);
  if(!(name.empty()) && (display_name.empty()))
    display_name = DA_ViewSpec::CleanName(name);
  TAPtr own = GetOwner();
  while((own != NULL) && (own->InheritsFrom(&TA_taList_impl))){
    own = own->GetOwner();
  }
  if(own != NULL){
    own->UpdateAfterEdit();
  }
}

void DT_ViewSpec::ButtonToggle(){
  visible = !visible;
  UpdateAfterEdit();
  char vstring = (visible ?  '1' : '0');
  taivMisc::RecordScript(GetPath() + ".visible = " + vstring + ";\n");
}


int DT_ViewSpec::BuildFromDataTable(DataTable* tdt){
  if(tdt != NULL) {
    if(data_table == NULL)
      taBase::SetPointer((TAPtr*)&data_table, tdt);
    if(name.empty()){
      name = tdt->name;
      display_name = DA_ViewSpec::CleanName(name);
    }
  }
  if(data_table == NULL)
    return false;

  // don't update if all sizes are the same..
  if((size == data_table->size) && (leaves == data_table->leaves) && (gp.size == data_table->gp.size))
    return false;

  // ensure number of elements is the same;
  if(data_table->size > size) {
    NewEl(data_table->size - size);
  }
  else {
    while(data_table->size < size){
      Remove(size-1);
    }
  }
  int i;
  for(i=0;i<size;i++) {
    DataArray_impl* nda =     ((DataArray_impl *) data_table->FastEl(i));
    DA_ViewSpec*    ndavs = (DA_ViewSpec *) FastEl(i);
    ndavs->BuildFromDataArray(nda);
  }


  // ensure the groups matchup
  if(data_table->gp.size > gp.size){
    NewGp(data_table->gp.size - gp.size,GetTypeDef());
  }
  else {
    while(data_table->gp.size < gp.size){
      gp.Remove(gp.size-1);
    }
  }
  for(i=0;i<data_table->gp.size;i++){
    DT_ViewSpec* nvs = (DT_ViewSpec *) Gp(i);
    DataTable* ndt = (DataTable *) data_table->Gp(i);
    nvs->el_base = nvs->el_typ = el_typ;
    nvs->BuildFromDataTable(ndt);
  }
  return true;
}

//////////////////////////////////
// 	DA Text View Specs	//
//////////////////////////////////

void DA_TextViewSpec::Initialize() {
  width = 2;
}

void DA_TextViewSpec::Destroy() {

}

int DA_TextViewSpec::BuildFromDataArray(DataArray_impl* tda) {
  int result = DA_ViewSpec::BuildFromDataArray(tda);
  if(data_array == NULL)	return result;
  if(data_array->HasDispOption(" NARROW, "))
    width = 1;
  String wd = data_array->DispOptionAfter(" WIDTH=");
  if(!wd.empty())
    width = (int)wd;
  return result;
}

//////////////////////////////////
// 	DA Page View Specs	//
//////////////////////////////////

void DA_PageViewSpec::Initialize() {
}

void DA_PageViewSpec::Destroy() {
}

int DA_PageViewSpec::BuildFromDataArray(DataArray_impl* tda) {
  int result = DA_ViewSpec::BuildFromDataArray(tda);
  if(data_array == NULL)	return result;
//  if(data_array->HasDispOption(" NARROW, "))
//    width = 1;
//  String wd = data_array->DispOptionAfter(" WIDTH=");
//  if(!wd.empty())
//    width = (int)wd;
  return result;
}

//////////////////////////////////
// 	DA Net View Specs	//
//////////////////////////////////

void DA_NetViewSpec::Initialize() {
  label_index = -1;
}

void DA_NetViewSpec::Destroy() {
}


//////////////////////////////////
// 	DA Graph View Specs	//
//////////////////////////////////

void DA_GraphViewSpec::Initialize(){
  line_type = LINE;
  line_style = SOLID;
  point_style = NONE;
  axis_spec = NULL;
  taBase::SetPointer((TAPtr*)&axis_spec,this);
  negative_draw = false;
  min_mode = AUTO_GROUP;
  max_mode = AUTO_GROUP;
  range.Init(0.0f);
  n_ticks = 10;
}

void DA_GraphViewSpec::Destroy() {
  CutLinks();
}

void DA_GraphViewSpec::CutLinks(){
  taBase::DelPointer((TAPtr*)&axis_spec);
}

void DA_GraphViewSpec::InitLinks(){
  DA_ViewSpec::InitLinks();
  taBase::Own(range, this);
  taBase::Own(line_color, this);
  // if it is in a subgroup then it will have 3 layers of DT_ViewSpec
  // above it. 3 becuase the first DT_ViewSpec is actual in a gp.
  // if it is in a subgroup then defaul to haveing the first elements graph
  DT_ViewSpec* myowner = GET_MY_OWNER(DT_ViewSpec);
  if(myowner != NULL) myowner = GET_OWNER(myowner,DT_ViewSpec);
  if(myowner != NULL) myowner = GET_OWNER(myowner,DT_ViewSpec);
  if(myowner != NULL) {
    myowner = GET_MY_OWNER(DT_ViewSpec);
    taBase::SetPointer((TAPtr*)&axis_spec,myowner->FastEl(0));
  }
}

void DA_GraphViewSpec::UpdateAfterEdit() {
  if(taMisc::iv_active) {
    ta_color.SetColor(line_color.r,line_color.g,line_color.b,
		      line_color.a);
  }
  if(axis_spec == NULL)		// don't like null axis specs..
    taBase::SetPointer((TAPtr*)&(axis_spec),this);
  DA_ViewSpec::UpdateAfterEdit();
}

void DA_GraphViewSpec::Copy_(const DA_GraphViewSpec& cp){
  line_color = cp.line_color;
  line_type = cp.line_type;
  line_style = cp.line_style;
  point_style = cp.point_style;
  point_mod = cp.point_mod;
  negative_draw = cp.negative_draw;
  taBase::SetPointer((TAPtr*)&axis_spec,this); // use this instead!
  min_mode = cp.min_mode;
  max_mode = cp.max_mode;
  range = cp.range;
  n_ticks = cp.n_ticks;
}

int DA_GraphViewSpec::BuildFromDataArray(DataArray_impl* tda) {
  int result = DA_ViewSpec::BuildFromDataArray(tda);
  if(data_array == NULL)	return result;
  if(data_array->InheritsFrom(TA_String_Data))
    visible = false;
  if(data_array->HasDispOption(" NARROW, "))
    visible = false;
  String val = data_array->DispOptionAfter(" MIN=");
  if(!val.empty()) {
    range.min = (float)val;
    range.UpdateAfterEdit();
    min_mode = FIXED;
  }
  val = data_array->DispOptionAfter(" MAX=");
  if(!val.empty()) {
    range.max = (float)val;
    range.UpdateAfterEdit();
    max_mode = FIXED;
  }
  return result;
}

//////////////////////////////////
// 	DT Graph View Specs	//
//////////////////////////////////

void DT_GraphViewSpec::Initialize() {
  background.name = "grey64";
  background.r = .6f;
  background.g = .6f;
  background.b = .6f;
  color_type = C_RAINBOW;
  sequence_1 = COLORS;
  sequence_2 = POINTS;
  sequence_3 = LINES;
}

void DT_GraphViewSpec::Destroy(){
}

void DT_GraphViewSpec::InitLinks() {
  DT_ViewSpec::InitLinks();
  taBase::Own(background, this);

  if(taMisc::iv_active) {
    bg_color.SetColor(background.r,background.g,background.b,background.a);
    ivWidgetKit* wkit = ivWidgetKit::instance();
    String guiname = wkit->gui();
    if(guiname == "monochrome") {
      color_type = M_MONO;
      sequence_1 = POINTS;
      sequence_2 = LINES;
      sequence_3 = COLORS;
    }
    else {
      color_type = C_RAINBOW;
      sequence_1 = COLORS;
      sequence_2 = POINTS;
      sequence_3 = LINES;
    }
  }	
}

void DT_GraphViewSpec::Copy_(const DT_GraphViewSpec& cp) {
  background = cp.background;
  color_type = cp.color_type;
  sequence_1 = cp.sequence_1;
  sequence_2 = cp.sequence_2;
  sequence_3 = cp.sequence_3;
}

int DT_GraphViewSpec::BuildFromDataTable(DataTable* tdt){
  int rval = DT_ViewSpec::BuildFromDataTable(tdt);
  if(rval == true)
    UpdateLineFeatures();
  return rval;
}

void DT_GraphViewSpec::UpdateAfterEdit() {
  if(taMisc::iv_active) {
    bg_color.SetColor(background.r,background.g,background.b,background.a);
  }
  DT_ViewSpec::UpdateAfterEdit();
}

static char* c_rainbow_colors[] = {
  "red", "orange", "yellow", "green", "blue", "navy", "violet",
  "brown", "black", "white", "pink", "pale green", "light blue" };
static int c_rainbow_count = 13;

static char* c_greyscale_colors[] = {
  "grey0", "grey10", "grey20", "grey30", "grey40", "grey60", "grey70", "grey80",
  "grey90", "grey100"};
static int c_greyscale_count = 10;

static char* m_mono_colors[] = { "black", "white" };
static int m_mono_count = 2;

static char* p_rainbow_colors[] = {
  "red", "orange", "yellow", "green", "blue", "navy", "violet",
  "brown", "black", "grey50", "pink", "pale green", "light blue" };
static int p_rainbow_count = 13;

static char* p_greyscale_colors[] = {
  "grey0", "grey10", "grey20", "grey30", "grey40", "grey50", "grey60", "grey70",
  "grey80", "grey90"};
static int p_greyscale_count = 10;

static char* p_mono_colors[] = { "black" };
static int p_mono_count = 1;


char* DT_GraphViewSpec::ColorName(int color_no) {
  switch(color_type) {
  case C_RAINBOW:
    return c_rainbow_colors[color_no % c_rainbow_count];
  case C_GREYSCALE:
    return c_greyscale_colors[color_no % c_greyscale_count];
  case M_MONO:
    return m_mono_colors[color_no % m_mono_count];
  case P_RAINBOW:
    return p_rainbow_colors[color_no % p_rainbow_count];
  case P_GREYSCALE:
    return p_greyscale_colors[color_no % p_greyscale_count];
  case P_MONO:
    return p_mono_colors[color_no % p_mono_count];
  case CUSTOM:
    return "";
  }
  return "";
}

int DT_GraphViewSpec::ColorCount() {
  switch(color_type) {
  case C_RAINBOW:
    return c_rainbow_count;
  case C_GREYSCALE:
    return c_greyscale_count;
  case M_MONO:
    return m_mono_count;
  case P_RAINBOW:
    return p_rainbow_count;
  case P_GREYSCALE:
    return p_greyscale_count;
  case P_MONO:
    return p_mono_count;
  case CUSTOM:
    return 0;
  }
  return 0;
}

void DT_GraphViewSpec::SetBgColor() {
  switch(color_type) {
  case C_RAINBOW:
  case C_GREYSCALE:
    background.name = "grey64";
    background.UpdateAfterEdit();
    break;
  case M_MONO:
    background.name = "white";
    background.UpdateAfterEdit();
    background.a = 1;
    break;
  case P_RAINBOW:
  case P_GREYSCALE:
  case P_MONO:
    background.name = "white";
    background.UpdateAfterEdit();
    break;
  case CUSTOM:
    break;
  }
  if(taMisc::iv_active)
    bg_color.SetColor(background.r,background.g,background.b,background.a);
}

void DT_GraphViewSpec::ApplyOneFeature(DA_GraphViewSpec* dagv, SequenceType seq,
				       int val, bool& set_ln, bool& set_pt) {
  if(seq == COLORS) {
    dagv->line_color.name = ColorName(val);
    dagv->line_color.UpdateAfterEdit();
  }
  else if(seq == LINES) {
    dagv->line_style = (DA_GraphViewSpec::LineStyle)val;
    set_ln = true;
  }
  else if(seq == POINTS) {
    dagv->point_style = (DA_GraphViewSpec::PointStyle)val;
    set_pt = true;
  }
}


void DT_GraphViewSpec::UpdateLineFeatures(bool visible_only) {
  int counts[4];
  counts[COLORS] = ColorCount();
  counts[LINES] = 4;
  counts[POINTS] = 4;
  counts[NONE] = 100000;	// should cover it..
  
  SetBgColor();

  int i=0;
  int cnt = 0;
  while(true) {
    int j = 0;
    while(true) {
      int k = 0;
      while(true) {
	DA_GraphViewSpec* dagv = (DA_GraphViewSpec*)Leaf(cnt++);
	if(dagv == NULL)
	  return;
	if((visible_only == true) && !dagv->visible)
	  continue;
	// initialize to default values first
	dagv->line_style = (DA_GraphViewSpec::LineStyle)0; 
	dagv->point_style = (DA_GraphViewSpec::PointStyle)0;
	bool set_ln = false;
	bool set_pt = false;
	ApplyOneFeature(dagv, sequence_1, k, set_ln, set_pt);
	ApplyOneFeature(dagv, sequence_2, j, set_ln, set_pt);
	ApplyOneFeature(dagv, sequence_3, i, set_ln, set_pt);

	if(set_ln && set_pt)
	  dagv->line_type = DA_GraphViewSpec::LINE_AND_POINTS;
	else if(set_pt) {
	  dagv->line_type = DA_GraphViewSpec::POINTS;
	  // add one to the point style since 0 is no point...
	  dagv->point_style = (DA_GraphViewSpec::PointStyle)
	    ((int)dagv->point_style + 1);
	}
	else			// default is line
	  dagv->line_type = DA_GraphViewSpec::LINE;
	

	dagv->UpdateAfterEdit();
	k++;
	if(k >= counts[sequence_1])			break;
      }
      j++;
      if(j >= counts[sequence_2])			break;
    }
    i++;
    if(i >= counts[sequence_3])				break;
  }
  
  UpdateAfterEdit();
}


//////////////////////////////////
// 	DA Grid View Specs	//
//////////////////////////////////

void DA_GridViewSpec::Initialize(){
  display_style = BLOCK;
}

void DA_GridViewSpec::Destroy() {

}

int DA_GridViewSpec::BuildFromDataArray(DataArray_impl* nda){
  int result = DA_ViewSpec::BuildFromDataArray(nda);
  if(data_array == NULL)
    return result;
  DT_GridViewSpec* ovs = GET_MY_OWNER(DT_GridViewSpec);
  if(ovs == NULL){
    pos.y = pos.x = 0;
  }
  else {
    if((ovs->geom.x == 0) && data_array->HasDispOption("GEOM_X")){
      ovs->geom.x = (int) data_array->DispOptionAfter("GEOM_X=");
    }
    if((ovs->geom.y == 0) && data_array->HasDispOption("GEOM_Y")){
      ovs->geom.y = (int) data_array->DispOptionAfter("GEOM_Y=");
    }
    if((pos.x == 0) && (pos.y == 0)) {
      ovs->GetPos(ovs->Find(this),pos.x, pos.y);
    }
  }
  if(data_array->HasDispOption("TEXT"))
    display_style = TEXT;		
  else if(data_array->HasDispOption("NARROW"))
    display_style = TEXT;		// use text for narrow items as a default..
  else if(data_array->InheritsFrom(TA_String_Data))
    display_style = TEXT;
  return result;
}

void DA_GridViewSpec::InitLinks(){
  DA_ViewSpec::InitLinks();
  taBase::Own(pos, this);
}

void DA_GridViewSpec::Copy_(const DA_GridViewSpec& cp) {
  pos = cp.pos;
  display_style = cp.display_style;
}


//////////////////////////////////
// 	DT Grid View Specs	//
//////////////////////////////////

void DT_GridViewSpec::Initialize(){
  use_gp_name = true;
  display_style = DA_GridViewSpec::BLOCK;
  ignore_overlap = false;
}

void DT_GridViewSpec::Destroy() {
}

void DT_GridViewSpec::InitLinks(){
  DT_ViewSpec::InitLinks();
  taBase::Own(geom, this);
  taBase::Own(pos, this);
}

void DT_GridViewSpec::Copy_(const DT_GridViewSpec& cp) {
  pos = cp.pos;
  geom = cp.geom;
  use_gp_name = cp.use_gp_name;
  display_style = cp.display_style;
}

int DT_GridViewSpec::BuildFromDataTable(DataTable* tdt){
  int result = DT_ViewSpec::BuildFromDataTable(tdt);
  if(data_table == NULL)
    return result;
  DT_GridViewSpec* ovs = GET_MY_OWNER(DT_GridViewSpec);
  if(ovs != NULL) {
    if((pos.x == 0) && (pos.y == 0)) {
      pos.x = ovs->MaxX() +1;
    }
  }
  if(result) UpdateGeom();
  return result;
}

void DT_GridViewSpec::UpdateGeom(){
  int i;
  for(i=0;i<size;i++){ // we only want the El's not the leaves
    DA_GridViewSpec* ar = (DA_GridViewSpec *) FastEl(i);
    geom.x = MAX(geom.x,ar->pos.x + 1);
    geom.y = MAX(geom.y,ar->pos.y + 1);
  }
}

void DT_GridViewSpec::UpdateAfterEdit(){
  ignore_overlap = false;
  if(CheckGeom() == true) UpdateLayout();
  DT_ViewSpec::UpdateAfterEdit();
}


int DT_GridViewSpec::CheckGeom(int xoff,int yoff,int mx, int my,int_Array* og){
  if(!visible) return false;
  int maxx;
  if(mx == -1) maxx = MaxX() +1;
  else maxx = mx;
  int maxy;
  if(my == -1) maxy = MaxY() +1;
  else maxy = my;
  int_Array* ogrid;
  if(og == NULL) {
    ogrid = new int_Array;
    ogrid->Insert(false,0,(maxx * maxy));
  }
  else ogrid = og;
  int i;
  for(i=0;i<size;i++){
    DA_GridViewSpec* ar = (DA_GridViewSpec*) FastEl(i);
    if(!ar->visible)
      continue;
    int oindex = ((ar->pos.y + pos.y + yoff) * maxx) + 
      (ar->pos.x + pos.x + xoff);
    if(!ignore_overlap && ((*ogrid)[oindex] == true)){
      String estring = String("Error:: Data overlap in grid layout for \"") +
	ar->display_name + "\"\n" + "at global position (" + 
	String(ar->pos.y + pos.y + yoff) +
	", " + String(ar->pos.x + pos.x + xoff) +")";
      int ch=taMisc::Choice(estring.chars(),"Reset Layout");
      if(ch==0)  return true; // fix the positions
    }
    (*ogrid)[oindex] = true;
  }
  int j;
  for(j=0;j<gp.size;j++) {
    DT_GridViewSpec* dtgvs = (DT_GridViewSpec*) gp.El(j);
    if(!dtgvs->visible)      continue;
    if(dtgvs->CheckGeom(pos.x + xoff,pos.y + yoff,maxx,maxy,ogrid) == true)
      return true;
  }
  if(og==NULL) delete ogrid;
  return false;
}



int DT_GridViewSpec::UpdateLayout(MatrixLayout ml){
  DA_GridViewSpec* da;
  int vis_size = 0;
  int i;
  for(i=0;i<size;i++){
    da = (DA_GridViewSpec *) FastEl(i);
    if(!da->visible) continue;
    vis_size++;
  }
  // insufficient geometry so use sqrt
  if((geom.x * geom.y) < vis_size) {
    if(vis_size < 25) {		// make it linear if a reasonable width..
      geom.x = vis_size; 
      geom.y = 1;
    }
    else {
      geom.y = (int)sqrt(vis_size);
      geom.x = vis_size / geom.y;
      while((geom.x * geom.y) < vis_size)
	geom.x++;
    }
  }
  int loc[2];
  int start[2];
  int mod[2];
  int limmiter = 0;
  int limit[2];  limit[0]= geom.x; limit[1] = geom.y;
  int maxx = -1;
  switch(ml){
  case LFT_RGT_TOP_BOT:
    loc[0] = start[0] = 0;    loc[1] = start[1] = 0;
    mod[0] = 1; mod[1] = 1; limmiter = 0;
    break;
  case LFT_RGT_BOT_TOP:
    loc[0] = start[0] = 0;    loc[1] = start[1] = geom.y-1;
    mod[0] = 1; mod[1] = -1; limmiter = 0;
    break;
  case BOT_TOP_LFT_RGT:
    loc[0] = start[0] = 0;    loc[1] = start[1] = geom.y-1;
    mod[0] = 1; mod[1] = -1;   limmiter = 1;
    break;
  case TOP_BOT_LFT_RGT:
    loc[0] = start[0] = 0;    loc[1] = start[1] = 0;
    mod[0] = 1; mod[1] = 1;   limmiter = 1;
    break;
  }
  for(i=0;i<size;i++){
    da = (DA_GridViewSpec *) FastEl(i);
    if(!da->visible) continue;
    da->pos.x = loc[0];
    da->pos.y = loc[1];
    maxx = MAX(maxx,loc[0]);
    loc[limmiter] += mod[limmiter];
    if(loc[limmiter] == limit[limmiter]){
      loc[limmiter] = start[limmiter];
      loc[!limmiter] += mod[!limmiter];
    }
  }
  loc[0] = maxx + 1;
  // now layout the subgroups to the right
  for(i=0;i<gp.size;i++){
    DT_GridViewSpec* dt = (DT_GridViewSpec*) gp.FastEl(i);
    if(!dt->visible) continue;
    dt->pos.x = loc[0];
    maxx = MAX(maxx,loc[0]);
    loc[0] += dt->UpdateLayout(ml) + 2;	// add a space between subsequent subgroups
  }
  return maxx;
}
				 

void	DT_GridViewSpec::GetPos(int apos, int&x, int&y, MatrixLayout ml){
  if(geom.x == 0) { x = apos; y = geom.y; return; }
  switch(ml){
  case LFT_RGT_TOP_BOT:
    y = (apos / geom.x);    x =  apos - (y * geom.x);    break;
  case LFT_RGT_BOT_TOP:
    y = (apos / geom.x);    x =  apos - (y * geom.x);    y = geom.y - y -1; break;
  case BOT_TOP_LFT_RGT:
    x = (apos / geom.y);    y =  apos - (x * geom.y);    break;
  case TOP_BOT_LFT_RGT:
    x = (apos / geom.y);    y =  apos - (x * geom.y);    x = geom.x - x -1; break;
  }
  if(y<0) y =0; // error?
  else if(y>geom.y-1) y = geom.y-1;
  if(x<0) x =0; // error?
  else if(x>geom.x-1) x = geom.x -1;
}

int DT_GridViewSpec::MaxY(){
  if(!visible) return 0;
  int max = pos.y;
  int i;
  DA_GridViewSpec* ar;
  for(i=0;i<size;i++){
    ar = (DA_GridViewSpec *) El(i);
    if(!ar->visible) continue;
    max = MAX(max,ar->pos.y + pos.y);
  }
  int j;
  for(j=0;j<gp.size;j++){
    DT_GridViewSpec* dt = (DT_GridViewSpec*) gp.El(j);
    if (!dt->visible) continue;
    max = MAX(max,(dt->MaxY() + pos.y));
  }
  return max;
}

int DT_GridViewSpec::MaxX(){
  if(!visible) return 0;
  int max = pos.x;
  int i;
  DA_GridViewSpec* ar;
  for(i=0;i<size;i++){
    ar = (DA_GridViewSpec *) El(i);
    if(!ar->visible) continue;
    max = MAX(max,ar->pos.x + pos.x);
  }
  int j;
  for(j=0;j<gp.size;j++){
    DT_GridViewSpec* dt = (DT_GridViewSpec*) gp.El(j);
    if (!dt->visible) continue;
    max = MAX(max,(dt->MaxX() + pos.x));
  }
  return max;
}


int DT_GridViewSpec::MaxHeaderY(){
  if(!visible) return 0;
  int max = pos.y;
  int i;
  DA_GridViewSpec* ar;
  for(i=0;i<size;i++){
    ar = (DA_GridViewSpec *) El(i);
    if(!ar->visible) continue;
    max = MAX(max,ar->pos.y + pos.y);
  }
  int j;
  for(j=0;j<gp.size;j++){
    DT_GridViewSpec* dt = (DT_GridViewSpec*) gp.El(j);
    if(!dt->visible) continue;
    if(dt->use_gp_name)
      max = MAX(max,dt->pos.y + pos.y);
    else
      max = MAX(max,dt->MaxY() + pos.y);
  }
  return max;
}

int DT_GridViewSpec::MaxHeaderX(){
  if(!visible) return 0;
  int max = pos.x;
  int i;
  DA_GridViewSpec* ar;
  for(i=0;i<size;i++){
    ar = (DA_GridViewSpec *) El(i);
    if(!ar->visible) continue;
    max = MAX(max,ar->pos.x + pos.x);
  }
  int j;
  for(j=0;j<gp.size;j++){
    DT_GridViewSpec* dt = (DT_GridViewSpec*) gp.El(j);
    if(!dt->visible) continue;
    if(dt->use_gp_name)
      max = MAX(max,dt->pos.x + pos.x);
    else
      max = MAX(max,dt->MaxX() + pos.x);
  }
  return max;
}

int DT_GridViewSpec::GetY(TAPtr ea){
  if(ea) {
    if(ea->InheritsFrom(&TA_DT_GridViewSpec))
      return (((DT_GridViewSpec*) ea)->pos.y);
    else {
      if(ea->InheritsFrom(&TA_DA_GridViewSpec)){
	return (((DA_GridViewSpec*) ea)->pos.y);
      }
    }
  }
  return 0;
}

int DT_GridViewSpec::GetX(TAPtr ea){
  if(ea) {
    if(ea->InheritsFrom(&TA_DT_GridViewSpec))
      return (((DT_GridViewSpec*) ea)->pos.x);
    else {
      if(ea->InheritsFrom(&TA_DA_GridViewSpec)){
	return (((DA_GridViewSpec*) ea)->pos.x);
      }
    }
  }
  return 0;
}

int DT_GridViewSpec::SetY(TAPtr ea, int y){
  if(!ea) return 0;

  if(ea->InheritsFrom(&TA_DT_GridViewSpec)) {
    ((DT_GridViewSpec*) ea)->pos.y = y;
    ((DT_GridViewSpec*) ea)->visible = true;
    return ((DT_GridViewSpec*) ea)->MaxY() -y +1;
  }
  else {
    if(ea->InheritsFrom(&TA_DA_GridViewSpec)){
      ((DA_GridViewSpec*) ea)->pos.y = y;
      ((DA_GridViewSpec*) ea)->visible = true;
      return 1;
    }
  }
  return 0;
}

int DT_GridViewSpec::SetX(TAPtr ea, int x){
  if(!ea) return 0;
  if(ea->InheritsFrom(&TA_DT_GridViewSpec)) {
    ((DT_GridViewSpec*) ea)->pos.x = x;
    ((DT_GridViewSpec*) ea)->visible = true;
    return ((DT_GridViewSpec*) ea)->MaxX() - x +1;
  }
  else {
    if(ea->InheritsFrom(&TA_DA_GridViewSpec)){
      ((DA_GridViewSpec*) ea)->pos.x = x;
      ((DA_GridViewSpec*) ea)->visible = true;
      return 1;
    }
  }
  return 0;
}
