/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#include <ta_misc/datatable_iv.h>

#include <ta/enter_iv.h>
#include <InterViews/font.h>
#include <IV-look/kit.h>
#include <IV-look/dialogs.h>
#include <InterViews/layout.h>
#include <InterViews/style.h>
#include <InterViews/background.h>
#include <InterViews/color.h>
#include <InterViews/stencil.h>
#include <InterViews/bitmap.h>
#include <InterViews/brush.h>
#include <InterViews/action.h>
#include <InterViews/group.h>
#include <InterViews/aggr.h>
#include <OS/math.h>
#include <ta/leave_iv.h>

#include <iv_misc/grid.h>
#include <iv_misc/hungrybox.h>
#include <iv_misc/vcrbitmaps.h>
#include <iv_misc/scrollable.h>
#include <iv_misc/dynalabel.h>
#include <iv_misc/dastepper.h>

////////////////////
//  DT_Action     //
////////////////////

// DT_Action is a special action class for when the user selects a headerlabel
// to be scaled on.

class DT_Action : public ivAction {
public:
  DT_CEdit_Obj* obj;
  ScaleBar*  sb;
  String name;
  InsetActiveButton* iab;
  HeaderLabel* hl;
  DT_Action(DT_CEdit_Obj* ob, ScaleBar* sb,char* nm);
  ~DT_Action();
  void execute();
};

DT_Action::DT_Action(DT_CEdit_Obj* ob,ScaleBar* s,char* nm){
  obj = ob;
  sb = s;
  ivResource::ref(sb);
  name = nm;
  iab = NULL;
}

DT_Action::~DT_Action() {
   ivResource::unref(sb);
}

void DT_Action::execute(){
  ScaleRange* sr = NULL;
  if(obj->GetScaleRange == NULL) return;
  obj->GetScaleRange(obj->scale_ctrl_obj,&sr);
  if((sr!=NULL) && (sr->grain == ScaleRange::FIXED)) { // set grain to BY_BUFFER if fixed
    sr->grain = ScaleRange::BY_BUFFER;
    sr->initial = sr->offset = 0.0f;
    taivMisc::RecordScript(sr->GetPath() + ".grain = BY_BUFFER;\n");
    taivMisc::RecordScript(sr->GetPath() + ".initial = 0.0;\n");
    taivMisc::RecordScript(sr->GetPath() + ".offset = 0.0;\n");
  }
  float srange = obj->GetScaleRange(obj->scale_ctrl_obj,NULL);
  sb->SetRange(srange);
  sb->UpdatePads();
  if(obj->ScriptScaleOn != NULL)
    obj->ScriptScaleOn(obj->scale_ctrl_obj,name,iab->down);
}
		      
//////////////////////////
//	HeaderLabel	//
//////////////////////////

void HeaderLabel::SetState(int toggle,int redraw){
  ivMonoGlyph* body1 = (ivMonoGlyph*) body();
  ivMonoGlyph* body2 = (ivMonoGlyph*) body1->body();
  InsetActiveButton* iab = (InsetActiveButton *) body2->body();
  ivTelltaleState* s = iab->state();
  if(toggle == -1)  // flip
    s->set(ivTelltaleState::is_active,!(s->test(ivTelltaleState::is_active)));
  else 
    s->set(ivTelltaleState::is_active,toggle);
  iab->down = s->test(ivTelltaleState::is_active);
  if(redraw == true) iab->redraw();
}
    

int HeaderLabel::GetTableIndex(DataTable** t, DT_GridViewSpec** gvs, int* n){
  *t = table; *n = index;   *gvs = dtgvs;
  if((index != -1 )&&
     ((table->El(index)->InheritsFrom(&TA_String_Data)) ||
      (((DA_GridViewSpec *) dtgvs->El(index))->display_style ==
	DA_GridViewSpec::TEXT))) return false;
  
  ivMonoGlyph* body1 = (ivMonoGlyph*) body();
  ivMonoGlyph* body2 = (ivMonoGlyph*) body1->body();
  InsetActiveButton* iab = (InsetActiveButton *) body2->body();

  return (iab->down);
}

String HeaderLabel::GetShortName(const String& visualname, float width) {
  ivWidgetKit* wkit = ivWidgetKit::instance();
  String shortname = visualname;
  ivFontBoundingBox fbb;
  wkit->font()->string_bbox(visualname.chars(),visualname.length(),fbb);
  int l = shortname.length();
  while(fbb.width() > width) {
    l--;
    shortname = shortname.through(l);
    wkit->font()->string_bbox(shortname.chars(),l,fbb);
  }
  return shortname;
}

#include <ta/enter_iv.h>
declareActionCallback(DT_ViewSpec)
implementActionCallback(DT_ViewSpec)
#include <ta/leave_iv.h>

#include <ta/enter_iv.h>
declareActionCallback(DA_ViewSpec)
implementActionCallback(DA_ViewSpec)
#include <ta/leave_iv.h>


HeaderLabel::HeaderLabel(DT_CEdit_Obj* dtobj, DataTable* t,
			 DT_GridViewSpec* gvs, int n)
: ivMonoGlyph(nil){
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivLayoutKit* layout = ivLayoutKit::instance();

  obj = dtobj;
  table = t;
  dtgvs = gvs;
  index = n;
  
  ivFontBoundingBox fbb;
  wkit->font()->font_bbox(fbb);
  ivCoord font_width = fbb.width();
  ivCoord font_height = wkit->font()->Height();
  ivGlyph* hlabel;
  ivTelltaleState* tt;


  ivCoord tablewidth = obj->GetBlockWidth();
  
  String visualname;		// what to display
  String shortname;		// short version for blocks

  if(index == -1){ // this is a data array object
    if(gvs->display_name.empty()) {
      if(!gvs->name.empty()) {
	gvs->display_name = gvs->name;
	visualname = gvs->name;
      }
      else  visualname = table->GetPath(table,table->GetOwner()->GetOwner());
    }
    else {
      visualname = gvs->display_name;
    }
    tablewidth *= (dtgvs->MaxX() - dtgvs->pos.x + 1);

    shortname = GetShortName(visualname, tablewidth);

    tt = new ivTelltaleState(ivTelltaleState::is_enabled);
    InsetActiveButton* iab;
    // advanced button bindings
    if(dtgvs->display_style >= DA_GridViewSpec::BLOCK) { // ok to scale on this one
      DT_Action* dta =  new DT_Action(obj,obj->hcsb,dtgvs->name);
      iab = new InsetActiveButton
	(wkit->push_button_look(wkit->label((char*) shortname), tt),
	 taivM->style,tt,
	 dta,
	 new ActionCallback(DT_ViewSpec) 
	 ((DT_ViewSpec *) dtgvs,&DT_ViewSpec::ButtonToggle),
	 new ActionCallback(DT_ViewSpec) 
	 ((DT_ViewSpec *) dtgvs,&DT_ViewSpec::ButtonEdit),
	 InsetActiveButton::BUTTON_FLIP,InsetActiveButton::BUTTON_SAME);
      dta->iab = iab;
    }
    else {
      iab = new InsetActiveButton
	(wkit->push_button_look(wkit->label((char*) visualname), tt),
	 taivM->style,tt,
	 NULL, // no action on left button
	 new ActionCallback(DT_ViewSpec) 
	 ((DT_ViewSpec *) dtgvs,&DT_ViewSpec::ButtonToggle),
	 new ActionCallback(DT_ViewSpec) 
	 ((DT_ViewSpec *) dtgvs,&DT_ViewSpec::ButtonEdit),
	 InsetActiveButton::BUTTON_UP,InsetActiveButton::BUTTON_SAME);
    }
    hlabel =  layout->hflexible
      (layout->natural (iab,tablewidth,font_height +4),0,0);
    body(hlabel);
    return;
  }

  // index != -1 thus this is an data array element not the array itself
  tt =  new ivTelltaleState(ivTelltaleState::is_enabled);
  
  DA_GridViewSpec* davs = (DA_GridViewSpec *) dtgvs->El(index);
  if(davs->display_name.empty()) {
    if(!davs->name.empty()) {
      davs->display_name = davs->name;
      visualname = davs->name;
    }
    else  visualname = table->GetPath(table,table->GetOwner()->GetOwner());
  }
  else {
    visualname = davs->display_name;
  }

  ivCoord headerwidth;
  if(davs->display_style == DA_GridViewSpec::TEXT) {
    if(table->FastEl(index)->HasDispOption(" NARROW, "))
      headerwidth = ((font_width * 4.0) + 4); // 4 for insetframe
    else
      headerwidth = (font_width * 8.0) + 4;
  }
  else
    headerwidth = tablewidth;

  shortname = GetShortName(visualname, headerwidth);

  InsetActiveButton* iab;
  if(davs->display_style >= DA_GridViewSpec::BLOCK) { // ok to scale on this one
    DT_Action* dta = new DT_Action(obj,obj->hcsb,davs->name);
    iab = new InsetActiveButton
      (wkit->push_button_look(wkit->label((char*) shortname), tt),
       taivM->style,tt,
       dta,
       new ActionCallback(DA_ViewSpec) 
       ((DA_ViewSpec *) davs,&DA_ViewSpec::ButtonToggle),
       new ActionCallback(DA_ViewSpec) 
       ((DA_ViewSpec *) davs,&DA_ViewSpec::ButtonEdit),
       InsetActiveButton::BUTTON_FLIP,InsetActiveButton::BUTTON_SAME);
    dta->iab = iab;
  }
  else {
    iab = new InsetActiveButton
      (wkit->push_button_look(wkit->label((char*) visualname), tt),
       taivM->style,tt,
       NULL,
       new ActionCallback(DA_ViewSpec) 
       ((DA_ViewSpec *) davs,&DA_ViewSpec::ButtonToggle),
       new ActionCallback(DA_ViewSpec) 
       ((DA_ViewSpec *) davs,&DA_ViewSpec::ButtonEdit),
       InsetActiveButton::BUTTON_UP,InsetActiveButton::BUTTON_SAME);
  }

  hlabel = 
    layout->hflexible
    (layout->natural(iab, headerwidth,font_height +4),0,0);
  body(hlabel);
}

//////////////////////////
//	DT_CEdit_Obj	//
//////////////////////////

DT_CEdit_Obj::~DT_CEdit_Obj(){
  ivResource::unref(headerbox);
  ivResource::unref(hcsb);
  taBase::DelPointer((TAPtr*)&cs);
}

void DT_CEdit_Obj::Initialize() {
  headerbox = NULL;
  eventbox= NULL;

  headergrid = NULL;
  eventpatch = NULL;
  print_patch = NULL;
  sb = NULL;

  scale_ctrl_obj = NULL;
  GetScaleOn = NULL;
  ScriptScaleOn = NULL;
  GetScaleRange = NULL;
  GetBlockSize = NULL;
  GetBlockBorder = NULL;

  fill_type = DT_GridViewSpec::COLOR;
  cs = NULL;
  taBase::OwnPointer((TAPtr*)&cs, new ColorScale, table);
  cs->NewDefaults();

  hcsb = NULL;
}

DT_CEdit_Obj::DT_CEdit_Obj(DataTable* dt,DT_GridViewSpec* gvs) {
  table = dt;
  dtgvs = gvs;
  Initialize();
}

int DT_CEdit_Obj::GetBlockWidth() {
  if((GetBlockSize == NULL) || (GetBlockBorder == NULL))
    return 14;
  return (GetBlockSize(scale_ctrl_obj) +
	  2 * GetBlockBorder(scale_ctrl_obj));
}

void  DT_CEdit_Obj::GetHeaderBox(DataTable* tdt,DT_GridViewSpec* gvs,
				 int xoff, int yoff)
{
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivLayoutKit* layout = ivLayoutKit::instance();
  if(tdt == NULL)    tdt = table;
  if(gvs == NULL)    gvs = dtgvs;
  if((gvs->size != tdt->size) || (gvs->gp.size != tdt->gp.size))
    gvs->BuildFromDataTable(tdt);

  int block_width = GetBlockWidth();

  static int_Array grid_widths;
  static int global_max_y;	// overall max y, x
  static int global_max_x;

  static ivGlyph* null_width;	// use this to fill grid with null width spacer
    
  if(tdt == table) {
    global_max_y = gvs->MaxHeaderY() + 1;
    global_max_x = gvs->MaxHeaderX() + 1;
    headergrid = new Grid(global_max_y, global_max_x);
    grid_widths.EnforceSize(global_max_y * global_max_x);
    
    ivGlyph* block_space = layout->hspace(block_width);
    null_width = layout->hspace(0);
    int x, y;
    int cnt = 0;
    for(y=0;y<global_max_y;y++) {
      for(x=0;x<global_max_x;x++, cnt++) {
 	headergrid->cell(y,x,block_space);
	grid_widths.FastEl(cnt) = -1; // -1 indicates empty grid location
      }
    }
    ivResource::unref(headerbox);
    headerbox = layout->hbox();
    ivResource::ref(headerbox);
  }

  // if gvs->use_gp_name then just use the group's name instead of the
  //  names of all the members.
  if(gvs->use_gp_name) {
    HeaderLabel* hlabel = new HeaderLabel(this, tdt, gvs);
    headerbox->append(hlabel);	// retain label ref
    int y = gvs->pos.y + yoff;
    int x = gvs->pos.x + xoff;
    headergrid->cell(y, x, hlabel);
    if(GetScaleOn != NULL)
      hlabel->SetState(GetScaleOn(scale_ctrl_obj,gvs->name),false);

    int maxx = gvs->MaxX();
    maxx = MIN(maxx,dtgvs->MaxHeaderX());
    grid_widths.FastEl(y * global_max_x + x) = maxx; // this spot now full
    // clear out extra grid slots to right of this extra-wide
    // label, which takes up more than its share of grid!
    int off = y * global_max_x;
    int i;
    for(i= x + 1; i <= maxx; i++) {
      headergrid->cell(y,i,null_width);
      grid_widths.FastEl(off + i) = 0;
    }
    // now clear out entire column for same width, but be sure not to
    // trample on anything that already lives there!
    int j;
    for(j=0; j<global_max_y; j++) {
      off = j * global_max_x;
      for(i= x + 1; i <= maxx; i++) {
	int& gw = grid_widths.FastEl(off + i);
	if(gw == -1) {		// only if currently empty!
	  headergrid->cell(j,i,null_width);
	  gw = 0;
	}
      }
    }
    return;
  }
  
  // otherwise do all the members.
  if(tdt->leaves == 0) {// nothing there, so put in header
    wkit->push_style(taivM->title_style);
    headergrid->cell
      (0,0, wkit->label
       ("-----------------------No-Current-Header------------------------"));
    wkit->pop_style();
    return;
  }

  if(gvs->visible == false) return;

  int i;
  for(i=0;i<tdt->size;i++){
    if(gvs->El(i) == NULL) continue;
    if(!(gvs->FastEl(i)->visible)) continue;
    HeaderLabel* hlabel = new HeaderLabel(this, tdt,gvs,i);
    DA_GridViewSpec* dagvs = (DA_GridViewSpec *) gvs->FastEl(i);
    headergrid->cell
      (dagvs->pos.y + gvs->pos.y + yoff, dagvs->pos.x + gvs->pos.x + xoff,
       hlabel);
    headerbox->append(hlabel);
    if(GetScaleOn != NULL)
      hlabel->SetState(GetScaleOn(scale_ctrl_obj,dagvs->name),false);
  }
  for(i=0;i<tdt->gp.size;i++){
    DataTable* dt  = ((DataTable *) tdt->gp.El(i));
    DT_GridViewSpec* gv = ((DT_GridViewSpec *) gvs->gp.El(i));
    if(gv->visible) GetHeaderBox(dt,gv,gvs->pos.x,gvs->pos.y);
  }
  ScaleRange* sr=NULL;
  if(GetScaleRange != NULL) GetScaleRange(scale_ctrl_obj,&sr);
  if((headerbox->count() == 1) && (sr->grain != ScaleRange::FIXED)) {
    // if only one label then select it
    HeaderLabel* hlabel = (HeaderLabel *) headerbox->component(0);
    hlabel->SetState(true);
  }
}

void DT_CEdit_Obj::FixSize(MinMax* view_range){
  if (!sb) return;
  GetEventBox(view_range);
}


ivGlyph* DT_CEdit_Obj::GetEventBox(MinMax* view_range,DataTable* tdt,
				   DT_GridViewSpec* gvs){
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivLayoutKit* layout = ivLayoutKit::instance();
  if(tdt == NULL) tdt = table;
  if(gvs == NULL) gvs = dtgvs;
  if((gvs->size != tdt->size) || (gvs->gp.size != tdt->gp.size))
    gvs->BuildFromDataTable(tdt);
  bool changed = false;
  int block_width = GetBlockWidth();
  ivGlyph* block_space = layout->hbox(layout->hspace(block_width),
				      layout->vspace(block_width));
  while(sb->count() < (view_range->max - view_range->min +1)) {
    int maxx = gvs->MaxX() +1;
    int maxy = gvs->MaxY() +1;
    ivResource::flush();
    Grid* grid = new Grid(maxy, maxx);
    int x, y;
    for(x=0;x<maxx;x++) {
      for(y=0;y<maxy;y++)
  	grid->cell(y,x,block_space);
    }
    headerindex = 0;
    ivGlyph* evnt = BuildEvent(0,grid,tdt,gvs);
    sb->append
      (layout->vbox
       (evnt,layout->hflexible(wkit->menu_item_separator_look())));
    events->modified(2);
    changed = true;
  }
  if(changed) eventpatch->reallocate();
  int i;
   for(i=(int)view_range->min;i<=view_range->max;i++) {
     headerindex = 0;
     SetEvent(i,(Grid *)
	      sb->component((ivGlyphIndex)(i-view_range->min))->component(0),tdt,gvs);
   }
  hcsb->SetRange(GetScaleRange(scale_ctrl_obj,NULL));
  ivResource::flush();
  return sb;
}

void DT_CEdit_Obj::SetBlockFill(DT_GridViewSpec::BlockFill b){
  if(fill_type == b) return;
  fill_type = b;
  int i;
  for(i=0;i<hcsb->padlist.size;i++){	
    ColorPad* cp = (ColorPad*) hcsb->padlist.FastEl(i);
    cp->SetFillType((ColorPad::BlockFill) b);
  }
}

ivGlyph* DT_CEdit_Obj::BuildEvent(int ev,Grid* egrid,
				DataTable* tdt, DT_GridViewSpec* gvs,
				  int xoff, int yoff ){
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivLayoutKit* layout = ivLayoutKit::instance();
  if(tdt == NULL) tdt = table;
  if(gvs == NULL) gvs = dtgvs;
  if((gvs->size != tdt->size) || (gvs->gp.size != tdt->gp.size))
    gvs->BuildFromDataTable(tdt);
  int i;
  ivGlyph* cellglyph;
  for(i=0;i<tdt->size;i++){
    DataArray_impl* da = tdt->FastEl(i);
    DA_GridViewSpec* dagvs = (DA_GridViewSpec *) gvs->FastEl(i);
    if(!(dagvs->visible)) continue;
    if((da->InheritsFrom(&TA_String_Data)) ||
       (gvs->display_style == DA_GridViewSpec::TEXT ) ||
       (dagvs->display_style == DA_GridViewSpec::TEXT)) {
      cellglyph = layout->vcenter
	(new DynamicLabel("XXXX",wkit->font(),wkit->foreground()),0);
    }
    else {
      String padname;
      if((gvs->display_style == DA_GridViewSpec::TEXT_AND_BLOCK) ||
	 (dagvs->display_style == DA_GridViewSpec::TEXT_AND_BLOCK))
	padname = "0";
      ColorPad* cp = new ColorPad(hcsb,(ColorPad::BlockFill) fill_type,
				  GetBlockSize(scale_ctrl_obj),GetBlockSize(scale_ctrl_obj),
				  padname);
      cellglyph = layout->margin(cp,GetBlockBorder(scale_ctrl_obj));
    }
    if(gvs->use_gp_name || (headerbox->count() <= headerindex)) {
      egrid->cell(dagvs->pos.y + gvs->pos.y + yoff
		  ,dagvs->pos.x + gvs->pos.x + xoff,
		  cellglyph);
    }
    else { // get the horizontal shape from the label
      egrid->cell(dagvs->pos.y + gvs->pos.y + yoff
		  ,dagvs->pos.x + gvs->pos.x + xoff,
		  new ShapeOfx
		  (layout->flexible
		   (layout->hbox
		    (layout->hglue(),cellglyph,layout->hglue()),
		    fil,0),
		   headerbox->component(headerindex++)));
    }
  }
  ivResource::flush();
  for(i=0;i<tdt->gp.size;i++){
    DataTable* dt  = ((DataTable *) tdt->gp.El(i));
    DT_GridViewSpec* dtg = (DT_GridViewSpec *) gvs->gp.El(i);
    if(!(dtg->visible)) continue;
    ivGlyph* result = BuildEvent(ev,egrid,dt,dtg,gvs->pos.x,gvs->pos.y);
    if(result == NULL) return NULL;
  }
  ivResource::flush();
  return egrid;
}

ivGlyph* DT_CEdit_Obj::SetEvent(int ev,Grid* egrid,
			      DataTable* tdt, DT_GridViewSpec* gvs,
				int xoff, int yoff ){
  if(tdt == NULL) tdt = table;
  if(gvs == NULL) gvs = dtgvs;
  if((gvs->size != tdt->size) || (gvs->gp.size != tdt->gp.size))
    gvs->BuildFromDataTable(tdt);
  int i;
  ivGlyph* cellglyph;
  int from_end = tdt->MaxLength() - ev -1;
  for(i=0;i<tdt->size;i++) {
    DataArray_impl* da = tdt->FastEl(i);
    DA_GridViewSpec* dagvs = (DA_GridViewSpec *) gvs->FastEl(i);
    if(!(dagvs->visible)) continue;
    if(gvs->use_gp_name || (headerbox->count() <= headerindex)) {
      cellglyph = egrid->component
	(dagvs->pos.y + gvs->pos.y + yoff,
	 dagvs->pos.x + gvs->pos.x + xoff);
    }
    else { // get the shape from the label
      cellglyph = ((ivMonoGlyph *) egrid->component
		   (dagvs->pos.y + gvs->pos.y + yoff
		    ,dagvs->pos.x + gvs->pos.x + xoff))->body();
      cellglyph = ((ivMonoGlyph *)cellglyph)->body()->component(1);
    }
    if(da->AR() == NULL)
      da->NewAR();
    if(cellglyph == NULL) continue;
    taArray_base* ar = da->AR();
    if((da->InheritsFrom(&TA_String_Data)) ||
       (gvs->display_style == DA_GridViewSpec::TEXT) ||
       (dagvs->display_style == DA_GridViewSpec::TEXT)) {
      String istr;
      if((ar->size > from_end) && (from_end >= 0))
	istr = ar->El_GetStr_(ar->FastEl_(ar->size-1-from_end));
      else 
	istr = "n/a";
      ((DynamicLabel *)((ivMonoGlyph *)cellglyph)->body())->set((char*)istr);
    }
    else {
      ColorPad* cp = (ColorPad *) ((ivMonoGlyph *) cellglyph)->body();
      if(cp == NULL)
	continue;
      float val;
      if((ar->size > from_end) && (from_end >= 0))
	val = *((float *)ar->FastEl_(ar->size-1-from_end));
      else val = 0.0f;
      cp->Set(val);
      if((gvs->display_style == DA_GridViewSpec::TEXT_AND_BLOCK) ||
	 (dagvs->display_style == DA_GridViewSpec::TEXT_AND_BLOCK)) 
	if(cp->name != String(val)) {
	  cp->name = String(val);
	  cp->ReFill();
	}
    }
  }
  ivResource::flush();
  for(i=0;i<tdt->gp.size;i++){
    DataTable* dt  = ((DataTable *) tdt->gp.El(i));
    DT_GridViewSpec* dtg = (DT_GridViewSpec *) gvs->gp.El(i);
    if(!(dtg->visible)) continue;
    SetEvent(ev,egrid,dt,dtg,gvs->pos.x,gvs->pos.y);
  }
  ivResource::flush();
  return egrid;
}


ivGlyph* DT_CEdit_Obj::Constr_Box(MinMax* view_range,
				int *feeder,
				int (*more)(int*,int, HungryBox*),
				int (*less)(int*,int, HungryBox*))
{
  ivWidgetKit* wkit = ivWidgetKit::instance();
  ivLayoutKit* layout = ivLayoutKit::instance();

  ivResource::unref(hcsb);
  hcsb = new HCScaleBar(0,true,true,new ivBrush(1),cs,16,8,18);
  ivResource::ref(hcsb);

  GetHeaderBox();
  sb = new TBHungryBox(feeder,more,less,8);
  eventbox = GetEventBox(view_range);
  eventpatch = new ivPatch(eventbox);

  events = layout->vbox
    (layout->hcenter(layout->vflexible(headergrid,nil,nil),0),
     layout->hflexible(wkit->menu_item_separator_look()),
     layout->hcenter(eventpatch,0));

  print_patch = new ivPatch(new ivBackground(events,cs->Get_Background()));

  ivGlyph* result = layout->vbox
    (layout->hcenter
     (layout->hbox(layout->hglue(),
		   wkit->inset_frame(print_patch),
		   layout->hglue()),0),
     layout->hcenter(hcsb->GetLook(),0));


  hcsb->UpdateScaleValues();

  return (new ivBackground(result,wkit->background()));
}




//////////////////////////////////
//	    DT_CEditDialog	//
//////////////////////////////////


DT_CEditDialog::DT_CEditDialog(DataTable* dt, DT_GridViewSpec* gvs) {
  table = dt;
  dtgvs = gvs;
  typ = dt->GetTypeDef();
  cur_base = dt;
  show = taMisc::show_iv;
  obj = new DT_CEdit_Obj(table,gvs);

  view_bufsz = MIN(3,dt->MinLength());
  view_range.min = 0;
  view_range.max = 2;
  view_shift = .2;
}

DT_CEditDialog::~DT_CEditDialog() {
  delete obj;
}

int DTCEDMore(int* tlv, int newsize, HungryBox*) {
 return ((DT_CEditDialog*) tlv)->ShowMore(newsize);
}

int DTCEDLess(int* tlv, int newsize, HungryBox*) {
  return ((DT_CEditDialog*) tlv)->ShowLess(newsize);
}

int DT_CEditDialog::ShowLess(int newsize){
  view_bufsz = MAX(newsize, 1);
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT((float) table->MinLength());
  if(view_range.max < 0) {
    view_range.max = -1;
    view_range.min = 0;
  }
  //  obj->FixSize(&view_range);
  return 0;
}

int DT_CEditDialog::ShowMore(int newsize){
  int oldsize = (int) (view_range.max - view_range.min +1);
  view_bufsz = MAX(newsize, 1);
  view_range.max = view_range.min + view_bufsz -1;
  view_range.MaxLT((float) table->MinLength());
  if(oldsize != (view_range.max - view_range.min + 1)) {
    //    obj->FixSize(&view_range);
    //    return 0;
    return 1;
  }
  else {
    return 1;
  }
}

void DT_CEditDialog::Constr_Box(){
  data = new ivInputHandler(obj->Constr_Box
			  (&view_range,(int *) this,
			   &DTCEDMore,&DTCEDLess),style);
  GetImage();
  data->append_input_handler(obj->hcsb->ihan);
  data->focus(obj->hcsb->ihan);
  box = wkit->inset_frame
    (new ivBackground
     (layout->vbox
      (layout->hbox(layout->hglue(),GetLogButtons(),layout->hglue()),
       layout->hflexible(wkit->menu_item_separator_look()),
       (ivGlyph*) data),
      wkit->background()->brightness
      (taivM->edit_darkbg_brightness)));
  // under some circumstances these are not used, so we ref them
  ivResource::ref(data);
  ivResource::ref(labels);
}

bool  DT_CEditDialog::ShowMember(MemberDef*) {
  return false;
}

void DT_CEditDialog::View_F() {
  int shft = (int)((float)view_bufsz * view_shift);
  if  (shft<1) shft = 1;
  view_range.max += shft;
  view_range.MaxLT(table->MinLength());
  view_range.MinGT(view_range.max - view_bufsz + 1);
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void DT_CEditDialog::View_FSF() {
  int shft = view_bufsz;
  if  (shft<1) shft = 1;
  view_range.max += shft;
  view_range.MaxLT(table->MinLength());
  view_range.MinGT(view_range.max - view_bufsz + 1);
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void DT_CEditDialog::View_R() {
  int shft = (int)((float)view_bufsz * view_shift);
  if  (shft<1) shft = 1;
  view_range.min -= shft;
  view_range.MinGT(0);
  view_range.MaxLT(view_range.min + view_bufsz -1);
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void DT_CEditDialog::View_FSR() {
  int shft = view_bufsz;
  if  (shft<1) shft = 1;
  view_range.min -= shft;
  view_range.MinGT(0);
  view_range.MaxLT(view_range.min + view_bufsz -1);
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void DT_CEditDialog::View_FF() {
  view_range.max = table->MinLength();
  view_range.MinGT(view_range.max - view_bufsz + 1); 
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

void DT_CEditDialog::View_FR() {
  view_range.min = 0;
  view_range.MaxLT(view_range.min + view_bufsz -1 );
  obj->FixSize(&view_range);
  obj->eventpatch->reallocate();
  obj->eventpatch->redraw();
}

#include <ta/enter_iv.h>
declareActionCallback(DT_CEditDialog)
implementActionCallback(DT_CEditDialog)
#include <ta/leave_iv.h>

ivGlyph* DT_CEditDialog::GetLogButtons() {
  ivStyle* run_style = new ivStyle(wkit->style());
  run_style->attribute("*PushButton*minimumWidth", "46");
  wkit->push_style(run_style);

  ivBitmap* fr_bm = new ivBitmap((void *) fr_bits,fr_width,fr_height);
  ivBitmap* fsr_bm = new ivBitmap((void *) fsr_bits,fsr_width,fsr_height);
  ivBitmap* r_bm = new ivBitmap((void *) r_bits,r_width,r_height);
  ivBitmap* f_bm = new ivBitmap((void *) f_bits,f_width,f_height);
  ivBitmap* fsf_bm = new ivBitmap((void *) fsf_bits,fsf_width,fsf_height);
  ivBitmap* ff_bm = new ivBitmap((void *) ff_bits,ff_width,ff_height);

  ivButton* frev = wkit->push_button(new ivStencil(fr_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_FR));
  ivButton* fsrev = wkit->push_button(new ivStencil(fsr_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_FSR));
  ivButton* rev = wkit->push_button(new ivStencil(r_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_R));
  ivButton* fwd = wkit->push_button(new ivStencil(f_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_F));
  ivButton* fsfwd = wkit->push_button(new ivStencil(fsf_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_FSF));
  ivButton* ffwd = wkit->push_button(new ivStencil(ff_bm,wkit->foreground()),
				 new ActionCallback(DT_CEditDialog)
				 (this,&DT_CEditDialog::View_FF));
  wkit->pop_style(); // deletes is too
  return (layout->hbox(frev,fsrev,rev,fwd,fsfwd,ffwd));
}






