/* -*- C++ -*- */
/*=============================================================================
//									      //
// This file is part of the TypeAccess/C-Super-Script software package.	      //
//									      //
// Copyright (C) 1995 Randall C. O'Reilly, Chadley K. Dawson, 		      //
//		      James L. McClelland, and Carnegie Mellon University     //
//     									      //
// Permission to use, copy, modify, and distribute this software and its      //
// documentation for any purpose is hereby granted without fee, provided that //
// the above copyright notice and this permission notice appear in all copies //
// of the software and related documentation.                                 //
// 									      //
// Note that the PDP++ software package, which contains this package, has a   //
// more restrictive copyright, which applies only to the PDP++-specific       //
// portions of the software, which are labeled as such.			      //
//									      //
// Note that the taString class, which is derived from the GNU String class,  //
// is Copyright (C) 1988 Free Software Foundation, written by Doug Lea, and   //
// is covered by the GNU General Public License, see ta_string.h.             //
// The iv_graphic library and some iv_misc classes were derived from the      //
// InterViews morpher example and other InterViews code, which is             //
// Copyright (C) 1987, 1988, 1989, 1990, 1991 Stanford University             //
// Copyright (C) 1991 Silicon Graphics, Inc.				      //
//									      //
// THE SOFTWARE IS PROVIDED "AS-IS" AND WITHOUT WARRANTY OF ANY KIND,         //
// EXPRESS, IMPLIED OR OTHERWISE, INCLUDING WITHOUT LIMITATION, ANY 	      //
// WARRANTY OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE.  	      //
// 									      //
// IN NO EVENT SHALL CARNEGIE MELLON UNIVERSITY BE LIABLE FOR ANY SPECIAL,    //
// INCIDENTAL, INDIRECT OR CONSEQUENTIAL DAMAGES OF ANY KIND, OR ANY DAMAGES  //
// WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER OR NOT     //
// ADVISED OF THE POSSIBILITY OF DAMAGE, AND ON ANY THEORY OF LIABILITY,      //
// ARISING OUT OF OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS        //
// SOFTWARE. 								      //
==============================================================================*/

#ifndef datatable_iv_h
#define datatable_iv_h

#include <ta/taiv_dialog.h>
#include <ta_misc/colorscale.h>
#include <ta_misc/colorbar.h>
#include <ta_misc/datatable.h>
#include <ta_misc/minmax.h>

class  	HCScaleBar; // #IGNORE
class 	Grid;	    // #IGNORE
class   HungryBox;  // #IGNORE
class   ivPatch;    // #IGNORE
class   ivMonoglyph;  // #IGNORE

class 	DT_CEdit_Obj;

class HeaderLabel : public ivMonoGlyph { // #IGNORE
public:
  DataTable*		table;
  DT_GridViewSpec*  	dtgvs;
  int			index;
  DT_CEdit_Obj* 	obj;

  String	GetShortName(const String& visualname, float width);
  int		GetTableIndex(DataTable** t, DT_GridViewSpec**  gvs, int* n);
  void  	SetState(int toggle =-1,int redraw=true); // -1 = flip, 0=off, 1=on

  HeaderLabel(DT_CEdit_Obj* dtobj, DataTable* t, DT_GridViewSpec* gvs, int n=-1);
};

class DT_CEdit_Obj { // #IGNORE
public:
  DT_GridViewSpec::BlockFill	fill_type;
  // the way that blocks are filled in the grid
  DataTable*	table;
  DT_GridViewSpec*  dtgvs;
  ColorScale*   cs;     
  HCScaleBar*	hcsb;
  ivGlyph*	headerbox;
  Grid*		headergrid;
  ivGlyph*	sb;	
  ivGlyph*	eventbox;
  ivPatch*	eventpatch;
  ivGlyph*	print_patch;
  ivPolyGlyph*	events;
  int		headerindex;

  void*		scale_ctrl_obj;	// object that controls scaling, other parameters
  // used as 'this' pointer for following functions, which control grid scaling
  bool		(*GetScaleOn)(void* obj, char* name);
  // returns true if given column is contributing to the scaling
  void		(*ScriptScaleOn)(void* obj, char* name, bool on_off);
  // record script code for turning scaling on or off for given column name
  float		(*GetScaleRange)(void* obj, ScaleRange** range);
  // if (range != NULL), fill in pointer to a ScaleRange object, else return scale range
  int		(*GetBlockSize)(void* obj);
  // return size (in points) of block to draw
  int		(*GetBlockBorder)(void* obj);
  // return size of border to draw around blocks

  int		GetBlockWidth(); // return total block width (size + 2*border)

  void GetHeaderBox(DataTable* tdt =NULL, DT_GridViewSpec* gvs=NULL,
		    int xoff =0, int yoff=0);
  ivGlyph* GetEventBox(MinMax* view_range,DataTable* tdt = NULL,
		       DT_GridViewSpec* gvs=NULL);
  ivGlyph* BuildEvent(int ev, Grid* egrid,
		      DataTable* tdt = NULL, DT_GridViewSpec* gvs=NULL,
		      int xoff=0, int yoff=0);
  ivGlyph* SetEvent(int ev, Grid* egrid, 
		    DataTable* tdt = NULL, DT_GridViewSpec* gvs=NULL,
		    int xoff=0, int yoff=0);
  ivGlyph* Constr_Box(MinMax* view_range, int *feeder,
		      int (*more)(int*,int, HungryBox*),
		      int (*less)(int*,int, HungryBox*));
  void  FixSize(MinMax* view_range);
  
  void SetBlockFill(DT_GridViewSpec::BlockFill b);

  void Initialize();
  DT_CEdit_Obj(DataTable* dt,DT_GridViewSpec* gvs);
  DT_CEdit_Obj(){Initialize();}
  ~DT_CEdit_Obj();
};

class DT_CEditDialog : public taivEditDialog {  // #NO_INSTANCE
public:
  DataTable*		table;
  DT_GridViewSpec*      dtgvs;
  DT_CEdit_Obj* 	obj;
  int 			labelindex;
  MinMax 		view_range;
  int			view_bufsz;
  float			view_shift;	// precentage of buffer to shift


  bool	ShowMember(MemberDef* md);
  void	Constr_Box();

  virtual void 	View_F();	// forward by view_shift
  virtual void 	View_R();	// rewind by view_shift
  virtual void 	View_FSF();	// forward by view_bufsz;
  virtual void 	View_FSR();	// rewind by view_bufsz;
  virtual void 	View_FF();	// forward to end
  virtual void 	View_FR();	// rewind to begining
  ivGlyph* GetLogButtons();

  int	ShowMore(int newsize);
  int	ShowLess(int newsize);

  DT_CEditDialog(DataTable* dt,DT_GridViewSpec* gvs);
  ~DT_CEditDialog();
  DT_CEditDialog() {};
};

#endif // datatable_iv_h



